#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES), and is copyright (c) 2018-2021
# by the software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia University
# Research Corporation, et al.  All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and
# license information.
#################################################################################

import os
import pytest
import idaes

class TestIdaesConfigure(object):

    @pytest.mark.unit
    def test_write_config(self):
        idaes.write_config(
            os.path.join(idaes.testing_directory, "config_testing_orig.json"))
        idaes.write_config(
            os.path.join(idaes.testing_directory, "config_testing_default.json"))


    @pytest.mark.unit
    def test_tmp_config_ctx(self):
        assert idaes.cfg["logging"]["disable_existing_loggers"] == False
        with idaes.temporary_config_ctx():
            idaes.cfg["logging"]["disable_existing_loggers"] = True
            assert idaes.cfg["logging"]["disable_existing_loggers"] == True
        assert idaes.cfg["logging"]["disable_existing_loggers"] == False


    @pytest.mark.unit
    def test_read_default_config(self):
        # set some stuff to not the default
        idaes.cfg.logging.disable_existing_loggers = True
        idaes.cfg.use_idaes_solvers = False
        assert idaes.cfg.logging.disable_existing_loggers == True
        assert idaes.cfg.use_idaes_solvers == False
        #read the default
        idaes.read_config(
            os.path.join(idaes.testing_directory, "config_testing_default.json"))
        assert idaes.cfg.ipopt.options.nlp_scaling_method == "gradient-based"
        assert idaes.cfg.use_idaes_solvers == True
        assert idaes.cfg.logging.disable_existing_loggers == False


    @pytest.mark.unit
    def test_change_env(self):
        with idaes.temporary_config_ctx():
            # default config uses idaes solvers
            idaes.cfg.use_idaes_solvers = False
            idaes.reconfig()
            pf = os.environ["PATH"]
        pt = os.environ["PATH"]
        assert pf != pt



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES), and is copyright (c) 2018-2021
# by the software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia University
# Research Corporation, et al.  All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and
# license information.
#################################################################################
"""
Author: Andrew Lee, Alejandro Garciadiego
"""
import pytest
from pyomo.environ import (ConcreteModel,
                           Set,
                           SolverStatus,
                           TerminationCondition,
                           value,
                           Var,
                           units as pyunits)
from pyomo.util.check_units import assert_units_consistent
from pyomo.common.unittest import assertStructuredAlmostEqual

from idaes.core import Component, FlowsheetBlock
from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.core.util.testing import initialization_tester
from idaes.core.util import get_solver


from idaes.generic_models.properties.core.generic.generic_property import (
        GenericParameterBlock)

from idaes.generic_models.unit_models import Flash
from idaes.generic_models.properties.core.state_definitions import FTPx
from idaes.generic_models.properties.core.phase_equil import SmoothVLE

from idaes.generic_models.properties.core.examples.CO2_bmimPF6_PR \
    import configuration


# -----------------------------------------------------------------------------
# Get default solver for testing
solver = get_solver()


# Test for configuration dictionaries with parameters from Properties of Gases
# and liquids 4th edition
class TestParamBlock(object):
    @pytest.mark.unit
    def test_build(self):
        model = ConcreteModel()

        model.param = GenericParameterBlock(default=configuration)

        assert isinstance(model.param.phase_list, Set)
        assert len(model.param.phase_list) == 2
        for i in model.param.phase_list:
            assert i in ["Liq", "Vap"]
        assert model.param.Liq.is_liquid_phase()
        assert model.param.Vap.is_vapor_phase()

        assert isinstance(model.param.component_list, Set)
        assert len(model.param.component_list) == 2
        for i in model.param.component_list:
            assert i in ['bmimPF6',
                         'carbon_dioxide']
            assert isinstance(model.param.get_component(i), Component)

        assert isinstance(model.param._phase_component_set, Set)
        assert len(model.param._phase_component_set) == 3
        for i in model.param._phase_component_set:
            assert i in [("Liq", "bmimPF6"), ("Liq", "carbon_dioxide"),
                         ("Vap", "carbon_dioxide")]

        assert model.param.config.state_definition == FTPx

        assertStructuredAlmostEqual(
            model.param.config.state_bounds,
            { "flow_mol": (0, 100, 1000, pyunits.mol/pyunits.s),
              "temperature": (10, 300, 500, pyunits.K),
              "pressure": (5e-4, 1e5, 1e10, pyunits.Pa) },
            item_callback=lambda x: value(x) * (
                pyunits.get_units(x) or pyunits.dimensionless)._get_pint_unit()
        )

        assert model.param.config.phase_equilibrium_state == {
            ("Vap", "Liq"): SmoothVLE}

        assert isinstance(model.param.phase_equilibrium_idx, Set)
        assert len(model.param.phase_equilibrium_idx) == 1
        for i in model.param.phase_equilibrium_idx:
            assert i in ["PE1"]

        assert model.param.phase_equilibrium_list == {
            "PE1": {"carbon_dioxide": ("Vap", "Liq")}}

        assert model.param.pressure_ref.value == 101325
        assert model.param.temperature_ref.value == 298.15

        assert model.param.bmimPF6.mw.value == 284.18E-3
        assert model.param.bmimPF6.pressure_crit.value == 24e5
        assert model.param.bmimPF6.temperature_crit.value == 860

        assert model.param.carbon_dioxide.mw.value == 44.010E-3
        assert model.param.carbon_dioxide.pressure_crit.value == 71.8e5
        assert model.param.carbon_dioxide.temperature_crit.value == 304.1

        assert_units_consistent(model)


class TestStateBlock(object):
    @pytest.fixture(scope="class")
    def model(self):
        model = ConcreteModel()

        model.fs = FlowsheetBlock(default={"dynamic": False})

        model.fs.param = GenericParameterBlock(default=configuration)

        model.fs.props = model.fs.param.build_state_block(
                [1],
                default={"defined_state": True})

        # Fix state
        model.fs.props[1].flow_mol.fix(1)
        model.fs.props[1].temperature.fix(298.15)
        model.fs.props[1].pressure.fix(1214713.75)
        model.fs.props[1].mole_frac_comp["carbon_dioxide"].fix(0.2)
        model.fs.props[1].mole_frac_comp["bmimPF6"].fix(0.8)

        assert degrees_of_freedom(model.fs.props[1]) == 0

        return model

    @pytest.mark.unit
    def test_build(self, model):
        # Check state variable values and bounds
        assert isinstance(model.fs.props[1].flow_mol, Var)
        assert value(model.fs.props[1].flow_mol) == 1
        assert model.fs.props[1].flow_mol.ub == 1000
        assert model.fs.props[1].flow_mol.lb == 0

        assert isinstance(model.fs.props[1].pressure, Var)
        assert value(model.fs.props[1].pressure) == 1214713.75
        assert model.fs.props[1].pressure.ub == 1e10
        assert model.fs.props[1].pressure.lb == 5e-4

        assert isinstance(model.fs.props[1].temperature, Var)
        assert value(model.fs.props[1].temperature) == 298.15
        assert model.fs.props[1].temperature.ub == 500
        assert model.fs.props[1].temperature.lb == 10

        assert isinstance(model.fs.props[1].mole_frac_comp, Var)
        assert len(model.fs.props[1].mole_frac_comp) == 2
        assert value(model.fs.props[1].mole_frac_comp["carbon_dioxide"]) == 0.2
        assert value(model.fs.props[1].mole_frac_comp["bmimPF6"]) == 0.8

        assert_units_consistent(model)

    @pytest.mark.unit
    def test_basic_scaling(self, model):
        model.fs.props[1].calculate_scaling_factors()

        assert len(model.fs.props[1].scaling_factor) == 18
        assert model.fs.props[1].scaling_factor[
            model.fs.props[1].flow_mol] == 1e-2
        assert model.fs.props[1].scaling_factor[
            model.fs.props[1].flow_mol_phase["Liq"]] == 1e-2
        assert model.fs.props[1].scaling_factor[
            model.fs.props[1].flow_mol_phase["Vap"]] == 1e-2
        assert model.fs.props[1].scaling_factor[
            model.fs.props[1].flow_mol_phase_comp[
                "Liq", "bmimPF6"]] == 1e-2
        assert model.fs.props[1].scaling_factor[
            model.fs.props[1].flow_mol_phase_comp[
                "Liq", "carbon_dioxide"]] == 1e-2
        assert model.fs.props[1].scaling_factor[
            model.fs.props[1].flow_mol_phase_comp[
                "Vap", "carbon_dioxide"]] == 1e-2
        assert model.fs.props[1].scaling_factor[
            model.fs.props[1].mole_frac_comp["bmimPF6"]] == 1000
        assert model.fs.props[1].scaling_factor[
            model.fs.props[1].mole_frac_comp["carbon_dioxide"]] == 1000
        assert model.fs.props[1].scaling_factor[
            model.fs.props[1].mole_frac_phase_comp["Liq", "bmimPF6"]] == 1000
        assert model.fs.props[1].scaling_factor[
            model.fs.props[1].mole_frac_phase_comp[
                "Liq", "carbon_dioxide"]] == 1000
        assert model.fs.props[1].scaling_factor[
            model.fs.props[1].mole_frac_phase_comp[
                "Vap", "carbon_dioxide"]] == 1000
        assert model.fs.props[1].scaling_factor[
            model.fs.props[1].pressure] == 1e-5
        assert model.fs.props[1].scaling_factor[
            model.fs.props[1].temperature] == 1e-2
        assert model.fs.props[1].scaling_factor[
            model.fs.props[1]._teq["Vap", "Liq"]] == 1e-2
        assert model.fs.props[1].scaling_factor[
            model.fs.props[1]._t1_Vap_Liq] == 1e-2

        assert model.fs.props[1].scaling_factor[
            model.fs.props[1]._mole_frac_tbub["Vap", "Liq", "bmimPF6"]] == 1
        assert model.fs.props[1].scaling_factor[
            model.fs.props[1]._mole_frac_tbub[
                "Vap", "Liq", "carbon_dioxide"]] == 1000
        assert model.fs.props[1].scaling_factor[
            model.fs.props[1].temperature_bubble["Vap", "Liq"]] == 1e-2

    @pytest.mark.unit
    def test_define_state_vars(self, model):
        sv = model.fs.props[1].define_state_vars()

        assert len(sv) == 4
        for i in sv:
            assert i in ["flow_mol",
                         "mole_frac_comp",
                         "temperature",
                         "pressure"]

    @pytest.mark.unit
    def test_define_port_members(self, model):
        sv = model.fs.props[1].define_state_vars()

        assert len(sv) == 4
        for i in sv:
            assert i in ["flow_mol",
                         "mole_frac_comp",
                         "temperature",
                         "pressure"]

    @pytest.mark.unit
    def test_define_display_vars(self, model):
        sv = model.fs.props[1].define_display_vars()

        assert len(sv) == 4
        for i in sv:
            assert i in ["Total Molar Flowrate",
                         "Total Mole Fraction",
                         "Temperature",
                         "Pressure"]

    @pytest.mark.ui
    @pytest.mark.unit
    def test_report(self, model):
        model.fs.props[1].report()


class TestFlashIntegration(object):
    @pytest.fixture(scope="class")
    def model(self):
        model = ConcreteModel()

        model.fs = FlowsheetBlock(default={"dynamic": False})

        model.fs.param = GenericParameterBlock(default=configuration)

        model.fs.unit = Flash(default={"property_package": model.fs.param,
                                       "has_heat_transfer": False,
                                       "has_pressure_change": False})
        # Fix state
        model.fs.unit.inlet.flow_mol.fix(1)
        model.fs.unit.inlet.temperature.fix(200.00)
        model.fs.unit.inlet.pressure.fix(101325)
        model.fs.unit.inlet.mole_frac_comp[0, "carbon_dioxide"].fix(1/2)
        model.fs.unit.inlet.mole_frac_comp[0, "bmimPF6"].fix(1/2)

        assert degrees_of_freedom(model.fs) == 0

        # Apply scaling - model will not solver without this
        model.fs.unit.control_volume.properties_in[
            0].calculate_scaling_factors()
        model.fs.unit.control_volume.properties_out[
            0].calculate_scaling_factors()

        return model

    @pytest.mark.component
    def test_initialize(self, model):

        initialization_tester(model, dof=0)

    @pytest.mark.component
    def test_solve(self, model):
        results = solver.solve(model)

        # Check for optimal solution
        assert results.solver.termination_condition == \
            TerminationCondition.optimal
        assert results.solver.status == SolverStatus.ok

    @pytest.mark.component
    def test_solution(self, model):
        # Check phase equilibrium results
        assert model.fs.unit.liq_outlet.mole_frac_comp[
            0, "carbon_dioxide"].value == \
            pytest.approx(0.3119, abs=1e-4)
        assert model.fs.unit.vap_outlet.mole_frac_comp[
            0, "carbon_dioxide"].value == \
            pytest.approx(1.0000, abs=1e-4)
        assert (model.fs.unit.vap_outlet.flow_mol[0].value /
                model.fs.unit.liq_outlet.flow_mol[0].value) == \
            pytest.approx(0.37619, abs=1e-4)

    @pytest.mark.ui
    @pytest.mark.unit
    def test_report(self, model):
        model.fs.unit.report()



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES), and is copyright (c) 2018-2021
# by the software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia University
# Research Corporation, et al.  All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and
# license information.
#################################################################################
"""
Methods for cubic equations of state.

Currently only supports liquid and vapor phases
"""
import os
from enum import Enum

from pyomo.environ import (exp,
                           Expression,
                           ExternalFunction,
                           log,
                           Param,
                           Reals,
                           sqrt,
                           Var)
from pyomo.common.config import ConfigBlock, ConfigValue, In

from idaes.core.util.exceptions import PropertyNotSupportedError
from idaes.generic_models.properties.core.generic.utility import (
    get_method, get_component_object as cobj)
from idaes.core.util.math import safe_log
from .eos_base import EoSBase
from idaes import bin_directory
import idaes.logger as idaeslog
from idaes.core.util.exceptions import \
    BurntToast, ConfigurationError, PropertyNotSupportedError


# Set up logger
_log = idaeslog.getLogger(__name__)


# Set path to root finder .so file
_so = os.path.join(bin_directory, "cubic_roots.so")


def cubic_roots_available():
    """Make sure the compiled cubic root functions are available. Yes, in
    Windows the .so extention is still used.
    """
    return os.path.isfile(_so)


class CubicType(Enum):
    PR = 0
    SRK = 1


class MixingRuleA(Enum):
    default = 0


class MixingRuleB(Enum):
    default = 0


EoS_param = {
        CubicType.PR: {'u': 2, 'w': -1, 'omegaA': 0.45724, 'coeff_b': 0.07780},
        CubicType.SRK: {'u': 1, 'w': 0, 'omegaA': 0.42748, 'coeff_b': 0.08664}
        }


CubicConfig = ConfigBlock()
CubicConfig.declare("type", ConfigValue(
    domain=In(CubicType),
    description="Equation of state to use",
    doc="Enum indicating type of cubic equation of state to use."))


class Cubic(EoSBase):

    @staticmethod
    def common(b, pobj):
        # TODO: determine if Henry's Law applies to Cubic EoS systems
        # For now, raise an exception if found
        # Follow on questions:
        # If Henry's law is used for a component, how does that effect
        # calculating A, B and phi?
        for j in b.component_list:
            cobj = b.params.get_component(j)
            if (cobj.config.henry_component is not None and
                    pobj.local_name in cobj.config.henry_component):
                raise PropertyNotSupportedError(
                    "{} Cubic equations of state do not support Henry's "
                    "components [{}, {}].".format(b.name, pobj.local_name, j))

        ctype = pobj._cubic_type
        cname = pobj.config.equation_of_state_options["type"].name

        if hasattr(b, cname+"_fw"):
            # Common components already constructed by previous phase
            return

        # Create expressions for coefficients
        def func_fw(m, j):
            cobj = m.params.get_component(j)
            if ctype == CubicType.PR:
                return 0.37464 + 1.54226*cobj.omega - \
                       0.26992*cobj.omega**2
            elif ctype == CubicType.SRK:
                return 0.48 + 1.574*cobj.omega - \
                       0.176*cobj.omega**2
            else:
                raise BurntToast(
                        "{} received unrecognized cubic type. This should "
                        "never happen, so please contact the IDAES developers "
                        "with this bug.".format(b.name))

        b.add_component(cname+'_fw',
                        Expression(b.component_list,
                                   rule=func_fw,
                                   doc='EoS S factor'))

        def func_a(m, j):
            cobj = m.params.get_component(j)
            fw = getattr(m, cname+"_fw")
            return (EoS_param[ctype]['omegaA']*(
                       (Cubic.gas_constant(b) *
                        cobj.temperature_crit)**2/cobj.pressure_crit) *
                    ((1+fw[j]*(1-sqrt(m.temperature /
                                      cobj.temperature_crit)))**2))
        b.add_component(cname+'_a',
                        Expression(b.component_list,
                                   rule=func_a,
                                   doc='Component a coefficient'))

        def func_b(m, j):
            cobj = m.params.get_component(j)
            return (EoS_param[ctype]['coeff_b'] * Cubic.gas_constant(b) *
                    cobj.temperature_crit/cobj.pressure_crit)
        b.add_component(cname+'_b',
                        Expression(b.component_list,
                                   rule=func_b,
                                   doc='Component b coefficient'))

        def rule_am(m, p):
            try:
                rule = m.params.get_phase(p).config.equation_of_state_options[
                    "mixing_rule_a"]
            except KeyError:
                rule = MixingRuleA.default

            a = getattr(m, cname+"_a")
            if rule == MixingRuleA.default:
                return rule_am_default(m, cname, a, p)
            else:
                raise ConfigurationError(
                    "{} Unrecognized option for Equation of State "
                    "mixing_rule_a: {}. Must be an instance of MixingRuleA "
                    "Enum.".format(m.name, rule))
        b.add_component(cname+'_am',
                        Expression(b.phase_list, rule=rule_am))

        def rule_bm(m, p):
            try:
                rule = m.params.get_phase(p).config.equation_of_state_options[
                    "mixing_rule_b"]
            except KeyError:
                rule = MixingRuleB.default

            b = getattr(m, cname+"_b")
            if rule == MixingRuleB.default:
                return rule_bm_default(m, b, p)
            else:
                raise ConfigurationError(
                    "{} Unrecognized option for Equation of State "
                    "mixing_rule_a: {}. Must be an instance of MixingRuleB "
                    "Enum.".format(m.name, rule))

        b.add_component(cname+'_bm',
                        Expression(b.phase_list, rule=rule_bm))

        def rule_A(m, p):
            am = getattr(m, cname+"_am")
            return (am[p]*m.pressure /
                    (Cubic.gas_constant(b)*m.temperature)**2)
        b.add_component(cname+'_A',
                        Expression(b.phase_list, rule=rule_A))

        def rule_B(m, p):
            bm = getattr(m, cname+"_bm")
            return (bm[p]*m.pressure /
                    (Cubic.gas_constant(b)*m.temperature))
        b.add_component(cname+'_B',
                        Expression(b.phase_list, rule=rule_B))

        def rule_delta(m, p, i):
            # See pg. 145 in Properties of Gases and Liquids
            a = getattr(m, cname+"_a")
            am = getattr(m, cname+"_am")
            kappa = getattr(m.params, cname+"_kappa")
            return (2*sqrt(a[i])/am[p] *
                    sum(m.mole_frac_phase_comp[p, j]*sqrt(a[j]) *
                        (1-kappa[i, j])
                        for j in b.components_in_phase(p)))
        b.add_component(cname+"_delta",
                        Expression(b.phase_component_set,
                                   rule=rule_delta))

        def rule_dadT(m, p):
            # See pg. 102 in Properties of Gases and Liquids
            a = getattr(m, cname+"_a")
            fw = getattr(m, cname+"_fw")
            kappa = getattr(m.params, cname+"_kappa")
            return -((Cubic.gas_constant(b)/2)*sqrt(EoS_param[ctype]['omegaA']) *
                     sum(sum(m.mole_frac_phase_comp[p, i] *
                             m.mole_frac_phase_comp[p, j] *
                             (1-kappa[i, j]) *
                             (fw[j]*sqrt(a[i] *
                              m.params.get_component(j).temperature_crit /
                              m.params.get_component(j).pressure_crit) +
                              fw[i]*sqrt(a[j] *
                              m.params.get_component(i).temperature_crit /
                              m.params.get_component(i).pressure_crit))
                             for j in m.components_in_phase(p))
                         for i in m.components_in_phase(p)) /
                     sqrt(m.temperature))
        b.add_component(cname+"_dadT",
                        Expression(b.phase_list,
                                   rule=rule_dadT))

        # Add components at equilibrium state if required
        if (b.params.config.phases_in_equilibrium is not None and
                (not b.config.defined_state or b.always_flash)):
            def func_a_eq(m, p1, p2, j):
                cobj = m.params.get_component(j)
                fw = getattr(m, cname+"_fw")
                return (EoS_param[ctype]['omegaA']*(
                            (Cubic.gas_constant(b) *
                             cobj.temperature_crit)**2/cobj.pressure_crit) *
                        ((1+fw[j]*(1-sqrt(m._teq[p1, p2] /
                                          cobj.temperature_crit)))**2))
            b.add_component('_'+cname+'_a_eq',
                            Expression(b.params._pe_pairs,
                                       b.component_list,
                                       rule=func_a_eq,
                                       doc='Component a coefficient at Teq'))

            def rule_am_eq(m, p1, p2, p3):
                try:
                    rule = m.params.get_phase(p3).config.equation_of_state_options[
                        "mixing_rule_a"]
                except KeyError:
                    rule = MixingRuleA.default

                a = getattr(m, "_"+cname+"_a_eq")
                if rule == MixingRuleA.default:
                    return rule_am_default(m, cname, a, p3, (p1, p2))
                else:
                    raise ConfigurationError(
                        "{} Unrecognized option for Equation of State "
                        "mixing_rule_a: {}. Must be an instance of MixingRuleA "
                        "Enum.".format(m.name, rule))
            b.add_component('_'+cname+'_am_eq',
                            Expression(b.params._pe_pairs,
                                       b.phase_list,
                                       rule=rule_am_eq))

            def rule_A_eq(m, p1, p2, p3):
                am_eq = getattr(m, "_"+cname+"_am_eq")
                return (am_eq[p1, p2, p3]*m.pressure /
                        (Cubic.gas_constant(b)*m._teq[p1, p2])**2)
            b.add_component('_'+cname+'_A_eq',
                            Expression(b.params._pe_pairs,
                                       b.phase_list,
                                       rule=rule_A_eq))

            def rule_B_eq(m, p1, p2, p3):
                bm = getattr(m, cname+"_bm")
                return (bm[p3]*m.pressure /
                        (Cubic.gas_constant(b)*m._teq[p1, p2]))
            b.add_component('_'+cname+'_B_eq',
                            Expression(b.params._pe_pairs,
                                       b.phase_list,
                                       rule=rule_B_eq))

            def rule_delta_eq(m, p1, p2, p3, i):
                # See pg. 145 in Properties of Gases and Liquids
                a = getattr(m, "_"+cname+"_a_eq")
                am = getattr(m, "_"+cname+"_am_eq")
                kappa = getattr(m.params, cname+"_kappa")
                return (2*sqrt(a[p1, p2, i])/am[p1, p2, p3] *
                        sum(m.mole_frac_phase_comp[p3, j]*sqrt(a[p1, p2, j]) *
                            (1-kappa[i, j])
                            for j in m.components_in_phase(p3)))
            b.add_component("_"+cname+"_delta_eq",
                            Expression(b.params._pe_pairs,
                                       b.phase_component_set,
                                       rule=rule_delta_eq))

        # Set up external function calls
        b.add_component("_"+cname+"_ext_func_param",
                        Param(default=ctype.value))
        b.add_component("_"+cname+"_proc_Z_liq",
                        ExternalFunction(library=_so,
                                         function="ceos_z_liq"))
        b.add_component("_"+cname+"_proc_Z_vap",
                        ExternalFunction(library=_so,
                                         function="ceos_z_vap"))

    @staticmethod
    def calculate_scaling_factors(b, pobj):
        pass

    @staticmethod
    def build_parameters(b):
        b._cubic_type = b.config.equation_of_state_options["type"]
        cname = b._cubic_type.name
        param_block = b.parent_block()

        if hasattr(param_block, cname+"_kappa"):
            # Common components already constructed by previous phase
            return

        kappa_data = param_block.config.parameter_data[cname+"_kappa"]
        param_block.add_component(
            cname+'_kappa',
            Var(param_block.component_list,
                param_block.component_list,
                within=Reals,
                initialize=kappa_data,
                doc=cname+' binary interaction parameters',
                units=None))

    @staticmethod
    def compress_fact_phase(b, p):
        pobj = b.params.get_phase(p)
        cname = pobj._cubic_type.name
        A = getattr(b, cname+"_A")
        B = getattr(b, cname+"_B")
        f = getattr(b, "_"+cname+"_ext_func_param")
        if pobj.is_vapor_phase():
            proc = getattr(b, "_"+cname+"_proc_Z_vap")
        elif pobj.is_liquid_phase():
            proc = getattr(b, "_"+cname+"_proc_Z_liq")
        else:
            raise PropertyNotSupportedError(_invalid_phase_msg(b.name, p))
        return proc(f, A[p], B[p])

    @staticmethod
    def dens_mass_phase(b, p):
        return b.dens_mol_phase[p]*b.mw_phase[p]

    @staticmethod
    def dens_mol_phase(b, p):
        pobj = b.params.get_phase(p)
        if pobj.is_vapor_phase() or pobj.is_liquid_phase():
            return b.pressure/(
                Cubic.gas_constant(b)*b.temperature*b.compress_fact_phase[p])
        else:
            raise PropertyNotSupportedError(_invalid_phase_msg(b.name, p))

    # TODO: Need to add functions to calculate cp and cv


    @staticmethod
    def energy_internal_mol_phase(blk, p):
        pobj = blk.params.get_phase(p)
        if not (pobj.is_vapor_phase() or pobj.is_liquid_phase()):
            raise PropertyNotSupportedError(_invalid_phase_msg(blk.name, p))

        cname = pobj._cubic_type.name
        am = getattr(blk, cname+"_am")[p]
        bm = getattr(blk, cname+"_bm")[p]
        B = getattr(blk, cname+"_B")[p]
        dadT = getattr(blk, cname+"_dadT")[p]
        Z = blk.compress_fact_phase[p]

        EoS_u = EoS_param[pobj._cubic_type]['u']
        EoS_w = EoS_param[pobj._cubic_type]['w']
        EoS_p = sqrt(EoS_u**2 - 4*EoS_w)

        # Derived from equation on pg. 120 in Properties of Gases and Liquids
        # Departure function for U is similar to H minus the RT(Z-1) term
        return (((blk.temperature*dadT - am) *
                 safe_log((2*Z + B*(EoS_u+EoS_p)) / (2*Z + B*(EoS_u-EoS_p)),
                          eps=1e-6)) / (bm*EoS_p) +
                sum(blk.mole_frac_phase_comp[p, j] *
                    EoSBase.energy_internal_mol_ig_comp_pure(blk, j)
                    for j in blk.components_in_phase(p)))

    @staticmethod
    def energy_internal_mol_phase_comp(blk, p, j):
        pobj = blk.params.get_phase(p)
        if not (pobj.is_vapor_phase() or pobj.is_liquid_phase()):
            raise PropertyNotSupportedError(_invalid_phase_msg(blk.name, p))

        cname = pobj._cubic_type.name
        am = getattr(blk, cname+"_am")[p]
        bm = getattr(blk, cname+"_bm")[p]
        B = getattr(blk, cname+"_B")[p]
        dadT = getattr(blk, cname+"_dadT")[p]
        Z = blk.compress_fact_phase[p]

        EoS_u = EoS_param[pobj._cubic_type]['u']
        EoS_w = EoS_param[pobj._cubic_type]['w']
        EoS_p = sqrt(EoS_u**2 - 4*EoS_w)

        # Derived from equation on pg. 120 in Properties of Gases and Liquids
        # Departure function for U is similar to H minus the RT(Z-1) term
        return (((blk.temperature*dadT - am) *
                 safe_log((2*Z + B*(EoS_u+EoS_p)) / (2*Z + B*(EoS_u-EoS_p)),
                          eps=1e-6)) / (bm*EoS_p) +
                EoSBase.energy_internal_mol_ig_comp_pure(blk, j))

    @staticmethod
    def enth_mol_phase(blk, p):
        pobj = blk.params.get_phase(p)
        if not (pobj.is_vapor_phase() or pobj.is_liquid_phase()):
            raise PropertyNotSupportedError(_invalid_phase_msg(blk.name, p))

        cname = pobj._cubic_type.name
        am = getattr(blk, cname+"_am")[p]
        bm = getattr(blk, cname+"_bm")[p]
        B = getattr(blk, cname+"_B")[p]
        dadT = getattr(blk, cname+"_dadT")[p]
        Z = blk.compress_fact_phase[p]

        EoS_u = EoS_param[pobj._cubic_type]['u']
        EoS_w = EoS_param[pobj._cubic_type]['w']
        EoS_p = sqrt(EoS_u**2 - 4*EoS_w)

        # Derived from equation on pg. 120 in Properties of Gases and Liquids
        return (((blk.temperature*dadT - am) *
                 safe_log((2*Z + B*(EoS_u+EoS_p)) / (2*Z + B*(EoS_u-EoS_p)),
                          eps=1e-6) +
                 Cubic.gas_constant(blk)*blk.temperature*(Z-1)*bm*EoS_p) /
                (bm*EoS_p) + sum(blk.mole_frac_phase_comp[p, j] *
                                 get_method(blk, "enth_mol_ig_comp", j)(
                                            blk, cobj(blk, j), blk.temperature)
                                 for j in blk.components_in_phase(p)))

    @staticmethod
    def enth_mol_phase_comp(blk, p, j):
        pobj = blk.params.get_phase(p)
        if not (pobj.is_vapor_phase() or pobj.is_liquid_phase()):
            raise PropertyNotSupportedError(_invalid_phase_msg(blk.name, p))

        cname = pobj._cubic_type.name
        am = getattr(blk, cname+"_am")[p]
        bm = getattr(blk, cname+"_bm")[p]
        B = getattr(blk, cname+"_B")[p]
        dadT = getattr(blk, cname+"_dadT")[p]
        Z = blk.compress_fact_phase[p]

        EoS_u = EoS_param[pobj._cubic_type]['u']
        EoS_w = EoS_param[pobj._cubic_type]['w']
        EoS_p = sqrt(EoS_u**2 - 4*EoS_w)

        # Derived from equation on pg. 120 in Properties of Gases and Liquids
        return (((blk.temperature*dadT - am) *
                 safe_log((2*Z + B*(EoS_u+EoS_p)) / (2*Z + B*(EoS_u-EoS_p)),
                          eps=1e-6) +
                 Cubic.gas_constant(blk)*blk.temperature*(Z-1)*bm*EoS_p) /
                (bm*EoS_p) + get_method(blk, "enth_mol_ig_comp", j)(
                                        blk, cobj(blk, j), blk.temperature))

    @staticmethod
    def entr_mol_phase(blk, p):
        pobj = blk.params.get_phase(p)
        if not (pobj.is_vapor_phase() or pobj.is_liquid_phase()):
            raise PropertyNotSupportedError(_invalid_phase_msg(blk.name, p))

        cname = pobj._cubic_type.name
        bm = getattr(blk, cname+"_bm")[p]
        B = getattr(blk, cname+"_B")[p]
        dadT = getattr(blk, cname+"_dadT")[p]
        Z = blk.compress_fact_phase[p]

        EoS_u = EoS_param[pobj._cubic_type]['u']
        EoS_w = EoS_param[pobj._cubic_type]['w']
        EoS_p = sqrt(EoS_u**2 - 4*EoS_w)

        # See pg. 102 in Properties of Gases and Liquids
        return ((Cubic.gas_constant(blk)*safe_log((Z-B)/Z, eps=1e-6)*bm*EoS_p +
                 Cubic.gas_constant(blk) *
                 safe_log(Z*blk.params.pressure_ref/blk.pressure, eps=1e-6) *
                 bm*EoS_p +
                 dadT*safe_log((2*Z + B*(EoS_u + EoS_p)) /
                               (2*Z + B*(EoS_u - EoS_p)),
                               eps=1e-6)) /
                (bm*EoS_p) + sum(blk.mole_frac_phase_comp[p, j] *
                                 get_method(blk, "entr_mol_ig_comp", j)(
                                     blk, cobj(blk, j), blk.temperature)
                                 for j in blk.components_in_phase(p)))

    @staticmethod
    def entr_mol_phase_comp(blk, p, j):
        pobj = blk.params.get_phase(p)
        if not (pobj.is_vapor_phase() or pobj.is_liquid_phase()):
            raise PropertyNotSupportedError(_invalid_phase_msg(blk.name, p))

        cname = pobj._cubic_type.name
        bm = getattr(blk, cname+"_bm")[p]
        B = getattr(blk, cname+"_B")[p]
        dadT = getattr(blk, cname+"_dadT")[p]
        Z = blk.compress_fact_phase[p]

        EoS_u = EoS_param[pobj._cubic_type]['u']
        EoS_w = EoS_param[pobj._cubic_type]['w']
        EoS_p = sqrt(EoS_u**2 - 4*EoS_w)

        # See pg. 102 in Properties of Gases and Liquids
        return (((Cubic.gas_constant(blk)*safe_log((Z-B)/Z, eps=1e-6) *
                  bm*EoS_p +
                  Cubic.gas_constant(blk) *
                  safe_log(Z*blk.params.pressure_ref /
                           (blk.mole_frac_phase_comp[p, j]*blk.pressure),
                           eps=1e-6) *
                  bm*EoS_p +
                  dadT*safe_log((2*Z + B*(EoS_u + EoS_p)) /
                                (2*Z + B*(EoS_u - EoS_p)),
                                eps=1e-6)) /
                (bm*EoS_p)) + get_method(blk, "entr_mol_ig_comp", j)(
                                      blk, cobj(blk, j), blk.temperature))

    @staticmethod
    def fug_phase_comp(b, p, j):
        pobj = b.params.get_phase(p)
        if pobj.is_vapor_phase() or pobj.is_liquid_phase():
            return (b.mole_frac_phase_comp[p, j] *
                    b.pressure *
                    b.fug_coeff_phase_comp[p, j])
        else:
            raise PropertyNotSupportedError(_invalid_phase_msg(b.name, p))

    @staticmethod
    def fug_phase_comp_eq(b, p, j, pp):
        pobj = b.params.get_phase(p)
        if pobj.is_vapor_phase() or pobj.is_liquid_phase():
            return (b.mole_frac_phase_comp[p, j] *
                    b.pressure *
                    exp(_log_fug_coeff_phase_comp_eq(b, p, j, pp)))
        else:
            raise PropertyNotSupportedError(_invalid_phase_msg(b.name, p))

    @staticmethod
    def log_fug_phase_comp_eq(b, p, j, pp):
        pobj = b.params.get_phase(p)
        if pobj.is_vapor_phase() or pobj.is_liquid_phase():
            return (log(b.mole_frac_phase_comp[p, j]) +
                    log(b.pressure/b.params.pressure_ref) +
                    _log_fug_coeff_phase_comp_eq(b, p, j, pp))
        else:
            raise PropertyNotSupportedError(_invalid_phase_msg(b.name, p))

    @staticmethod
    def fug_coeff_phase_comp(blk, p, j):
        pobj = blk.params.get_phase(p)
        ctype = pobj._cubic_type

        if not (pobj.is_vapor_phase() or pobj.is_liquid_phase()):
            raise PropertyNotSupportedError(_invalid_phase_msg(blk.name, p))

        cname = pobj._cubic_type.name
        b = getattr(blk, cname+"_b")[j]
        bm = getattr(blk, cname+"_bm")[p]
        A = getattr(blk, cname+"_A")[p]
        B = getattr(blk, cname+"_B")[p]
        delta = getattr(blk, cname+"_delta")[p, j]
        Z = blk.compress_fact_phase[p]

        return exp(_log_fug_coeff_method(A, b, bm, B, delta, Z, ctype))

    @staticmethod
    def fug_coeff_phase_comp_eq(blk, p, j, pp):
        return exp(_log_fug_coeff_phase_comp_eq(blk, p, j, pp))

    @staticmethod
    def log_fug_phase_comp_Tbub(blk, p, j, pp):
        pobj = blk.params.get_phase(p)
        ctype = pobj._cubic_type
        cname = pobj.config.equation_of_state_options["type"].name

        if pobj.is_liquid_phase():
            x = blk.mole_frac_comp
            xidx = ()
        elif pobj.is_vapor_phase():
            x = blk._mole_frac_tbub
            xidx = pp
        else:
            raise BurntToast("{} non-vapor or liquid phase called for bubble "
                             "temperature calculation. This should never "
                             "happen, so please contact the IDAES developers "
                             "with this bug.".format(blk.name))

        def a(k):
            cobj = blk.params.get_component(k)
            fw = getattr(blk, cname+"_fw")[k]
            return (EoS_param[ctype]['omegaA'] *
                    ((Cubic.gas_constant(blk) * cobj.temperature_crit)**2 /
                     cobj.pressure_crit) *
                    ((1+fw*(1-sqrt(blk.temperature_bubble[pp] /
                                   cobj.temperature_crit)))**2))

        kappa = getattr(blk.params, cname+"_kappa")
        am = sum(sum(x[xidx, i]*x[xidx, j]*sqrt(a(i)*a(j))*(1-kappa[i, j])
                     for j in blk.component_list)
                 for i in blk.component_list)

        b = getattr(blk, cname+"_b")
        bm = sum(x[xidx, i]*b[i] for i in blk.component_list)

        A = am*blk.pressure/(Cubic.gas_constant(blk) *
                             blk.temperature_bubble[pp])**2
        B = bm*blk.pressure/(Cubic.gas_constant(blk) *
                             blk.temperature_bubble[pp])

        delta = (2*sqrt(a(j))/am * sum(x[xidx, i]*sqrt(a(i))*(1-kappa[j, i])
                                       for i in blk.component_list))

        f = getattr(blk, "_"+cname+"_ext_func_param")
        if pobj.is_vapor_phase():
            proc = getattr(blk, "_"+cname+"_proc_Z_vap")
        elif pobj.is_liquid_phase():
            proc = getattr(blk, "_"+cname+"_proc_Z_liq")

        Z = proc(f, A, B)

        if pobj.is_vapor_phase():
            mole_frac = blk._mole_frac_tbub[pp[0], pp[1], j]
        else:
            mole_frac = blk.mole_frac_comp[j]

        return (_log_fug_coeff_method(A, b[j], bm, B, delta, Z, ctype) +
                log(mole_frac) + log(blk.pressure/blk.pressure._units))

    @staticmethod
    def log_fug_phase_comp_Tdew(blk, p, j, pp):
        pobj = blk.params.get_phase(p)
        ctype = pobj._cubic_type
        cname = pobj.config.equation_of_state_options["type"].name

        if pobj.is_liquid_phase():
            x = blk._mole_frac_tdew
            xidx = pp
        elif pobj.is_vapor_phase():
            x = blk.mole_frac_comp
            xidx = ()
        else:
            raise BurntToast("{} non-vapor or liquid phase called for bubble "
                             "temperature calculation. This should never "
                             "happen, so please contact the IDAES developers "
                             "with this bug.".format(blk.name))

        def a(k):
            cobj = blk.params.get_component(k)
            fw = getattr(blk, cname+"_fw")[k]
            return (EoS_param[ctype]['omegaA'] *
                    ((Cubic.gas_constant(blk) * cobj.temperature_crit)**2 /
                     cobj.pressure_crit) *
                    ((1+fw*(1-sqrt(blk.temperature_dew[pp] /
                                   cobj.temperature_crit)))**2))

        kappa = getattr(blk.params, cname+"_kappa")
        am = sum(sum(x[xidx, i]*x[xidx, j]*sqrt(a(i)*a(j))*(1-kappa[i, j])
                     for j in blk.component_list)
                 for i in blk.component_list)

        b = getattr(blk, cname+"_b")
        bm = sum(x[xidx, i]*b[i] for i in blk.component_list)

        A = am*blk.pressure/(Cubic.gas_constant(blk) *
                             blk.temperature_dew[pp])**2
        B = bm*blk.pressure/(Cubic.gas_constant(blk) *
                             blk.temperature_dew[pp])

        delta = (2*sqrt(a(j))/am * sum(x[xidx, i]*sqrt(a(i))*(1-kappa[j, i])
                                       for i in blk.component_list))

        f = getattr(blk, "_"+cname+"_ext_func_param")
        if pobj.is_vapor_phase():
            proc = getattr(blk, "_"+cname+"_proc_Z_vap")
        elif pobj.is_liquid_phase():
            proc = getattr(blk, "_"+cname+"_proc_Z_liq")

        Z = proc(f, A, B)

        if pobj.is_vapor_phase():
            mole_frac = blk.mole_frac_comp[j]
        else:
            mole_frac = blk._mole_frac_tdew[pp[0], pp[1], j]

        return (_log_fug_coeff_method(A, b[j], bm, B, delta, Z, ctype) +
                log(mole_frac) + log(blk.pressure/blk.pressure._units))

    @staticmethod
    def log_fug_phase_comp_Pbub(blk, p, j, pp):
        pobj = blk.params.get_phase(p)
        ctype = pobj._cubic_type
        cname = pobj.config.equation_of_state_options["type"].name

        if pobj.is_liquid_phase():
            x = blk.mole_frac_comp
            xidx = ()
        elif pobj.is_vapor_phase():
            x = blk._mole_frac_pbub
            xidx = pp
        else:
            raise BurntToast("{} non-vapor or liquid phase called for bubble "
                             "temperature calculation. This should never "
                             "happen, so please contact the IDAES developers "
                             "with this bug.".format(blk.name))

        a = getattr(blk, cname+"_a")
        kappa = getattr(blk.params, cname+"_kappa")
        am = sum(sum(x[xidx, i]*x[xidx, j] *
                     sqrt(a[i]*a[j])*(1-kappa[i, j])
                     for j in blk.component_list)
                 for i in blk.component_list)

        b = getattr(blk, cname+"_b")
        bm = sum(x[xidx, i]*b[i] for i in blk.component_list)

        A = am*blk.pressure_bubble[pp]/(Cubic.gas_constant(blk) *
                                        blk.temperature)**2
        B = bm*blk.pressure_bubble[pp]/(Cubic.gas_constant(blk) *
                                        blk.temperature)

        delta = (2*sqrt(a[j])/am * sum(x[xidx, i]*sqrt(a[i])*(1-kappa[j, i])
                                       for i in blk.component_list))

        f = getattr(blk, "_"+cname+"_ext_func_param")
        if pobj.is_vapor_phase():
            proc = getattr(blk, "_"+cname+"_proc_Z_vap")
        elif pobj.is_liquid_phase():
            proc = getattr(blk, "_"+cname+"_proc_Z_liq")

        Z = proc(f, A, B)

        if pobj.is_vapor_phase():
            mole_frac = blk._mole_frac_pbub[pp[0], pp[1], j]
        else:
            mole_frac = blk.mole_frac_comp[j]

        return (_log_fug_coeff_method(A, b[j], bm, B, delta, Z, ctype) +
                log(mole_frac) + log(blk.pressure_bubble[pp] /
                                     blk.pressure_bubble._units))

    @staticmethod
    def log_fug_phase_comp_Pdew(blk, p, j, pp):
        pobj = blk.params.get_phase(p)
        ctype = pobj._cubic_type
        cname = pobj.config.equation_of_state_options["type"].name

        if pobj.is_liquid_phase():
            x = blk._mole_frac_pdew
            xidx = pp
        elif pobj.is_vapor_phase():
            x = blk.mole_frac_comp
            xidx = ()
        else:
            raise BurntToast("{} non-vapor or liquid phase called for bubble "
                             "temperature calculation. This should never "
                             "happen, so please contact the IDAES developers "
                             "with this bug.".format(blk.name))

        a = getattr(blk, cname+"_a")
        kappa = getattr(blk.params, cname+"_kappa")
        am = sum(sum(x[xidx, i]*x[xidx, j] *
                     sqrt(a[i]*a[j])*(1-kappa[i, j])
                     for j in blk.component_list)
                 for i in blk.component_list)

        b = getattr(blk, cname+"_b")
        bm = sum(x[xidx, i]*b[i] for i in blk.component_list)

        A = am*blk.pressure_dew[pp]/(Cubic.gas_constant(blk) *
                                     blk.temperature)**2
        B = bm*blk.pressure_dew[pp]/(Cubic.gas_constant(blk)*blk.temperature)

        delta = (2*sqrt(a[j])/am * sum(x[xidx, i]*sqrt(a[i])*(1-kappa[j, i])
                                       for i in blk.component_list))

        f = getattr(blk, "_"+cname+"_ext_func_param")
        if pobj.is_vapor_phase():
            proc = getattr(blk, "_"+cname+"_proc_Z_vap")
        elif pobj.is_liquid_phase():
            proc = getattr(blk, "_"+cname+"_proc_Z_liq")

        Z = proc(f, A, B)

        if pobj.is_vapor_phase():
            mole_frac = blk.mole_frac_comp[j]
        else:
            mole_frac = blk._mole_frac_pdew[pp[0], pp[1], j]

        return (_log_fug_coeff_method(A, b[j], bm, B, delta, Z, ctype) +
                log(mole_frac) + log(blk.pressure_dew[pp] /
                                     blk.pressure_dew._units))

    @staticmethod
    def gibbs_mol_phase(b, p):
        return (b.enth_mol_phase[p] - b.entr_mol_phase[p]*b.temperature)

    @staticmethod
    def gibbs_mol_phase_comp(b, p, j):
        return (b.enth_mol_phase_comp[p, j] -
                b.entr_mol_phase_comp[p, j] *
                b.temperature)

    @staticmethod
    def vol_mol_phase(b, p):
        pobj = b.params.get_phase(p)
        if pobj.is_vapor_phase() or pobj.is_liquid_phase():
            return (Cubic.gas_constant(b)*b.temperature *
                    b.compress_fact_phase[p] /
                    b.pressure)
        else:
            raise PropertyNotSupportedError(_invalid_phase_msg(b.name, p))


def _invalid_phase_msg(name, phase):
    return ("{} received unrecognized phase name {}. Ideal property "
            "libray only supports Vap and Liq phases."
            .format(name, phase))


def _log_fug_coeff_phase_comp_eq(blk, p, j, pp):
    pobj = blk.params.get_phase(p)
    if not (pobj.is_vapor_phase() or pobj.is_liquid_phase()):
        raise PropertyNotSupportedError(_invalid_phase_msg(blk.name, p))

    cname = pobj._cubic_type.name
    b = getattr(blk, cname+"_b")
    bm = getattr(blk, cname+"_bm")
    Aeq = getattr(blk, "_"+cname+"_A_eq")
    Beq = getattr(blk, "_"+cname+"_B_eq")
    delta_eq = getattr(blk, "_"+cname+"_delta_eq")

    f = getattr(blk, "_"+cname+"_ext_func_param")
    if pobj.is_vapor_phase():
        proc = getattr(blk, "_"+cname+"_proc_Z_vap")
    elif pobj.is_liquid_phase():
        proc = getattr(blk, "_"+cname+"_proc_Z_liq")

    def Zeq(p):
        return proc(f, Aeq[pp, p], Beq[pp, p])

    return _log_fug_coeff_method(Aeq[pp, p], b[j], bm[p], Beq[pp, p],
                                 delta_eq[pp, p, j], Zeq(p), pobj._cubic_type)


def _log_fug_coeff_method(A, b, bm, B, delta, Z, cubic_type):
    u = EoS_param[cubic_type]['u']
    w = EoS_param[cubic_type]['w']
    p = sqrt(u**2 - 4*w)

    return ((b/bm*(Z-1)*(B*p) - safe_log(Z-B, eps=1e-6)*(B*p) +
             A*(b/bm - delta)*safe_log((2*Z + B*(u + p))/(2*Z + B*(u - p)),
                                       eps=1e-6)) /
            (B*p))


# -----------------------------------------------------------------------------
# Mixing rules
def rule_am_default(m, cname, a, p, pp=()):
    k = getattr(m.params, cname+"_kappa")
    return sum(sum(
        m.mole_frac_phase_comp[p, i]*m.mole_frac_phase_comp[p, j] *
        sqrt(a[pp, i]*a[pp, j])*(1-k[i, j])
        for j in m.components_in_phase(p))
        for i in m.components_in_phase(p))


def rule_bm_default(m, b, p):
    return sum(m.mole_frac_phase_comp[p, i]*b[i]
               for i in m.components_in_phase(p))



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES), and is copyright (c) 2018-2021
# by the software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia University
# Research Corporation, et al.  All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and
# license information.
#################################################################################
from .process_base import ProcessBlockData, useDefault, MaterialFlowBasis
from .process_block import ProcessBlock, declare_process_block_class
from .unit_model import UnitModelBlockData, UnitModelBlock
from .flowsheet_model import FlowsheetBlockData, FlowsheetBlock
from .property_base import (StateBlockData, PhysicalParameterBlock,
                            StateBlock)
from .reaction_base import (ReactionBlockDataBase, ReactionParameterBlock,
                            ReactionBlockBase)
from .control_volume_base import (ControlVolumeBlockData, CONFIG_Template,
                                  MaterialBalanceType, EnergyBalanceType,
                                  MomentumBalanceType, FlowDirection)
from .control_volume0d import ControlVolume0DBlock
from .control_volume1d import ControlVolume1DBlock
from .phases import (
    Phase, LiquidPhase, SolidPhase, VaporPhase, PhaseType, AqueousPhase)
from .components import Component, Solvent, Solute, Ion, Cation, Anion, Apparent



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES), and is copyright (c) 2018-2021
# by the software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia University
# Research Corporation, et al.  All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and
# license information.
#################################################################################
"""
Tests for Reaction Interrogator Tool

@author: alee
"""
import pytest

from pyomo.environ import ConcreteModel, units as pyunits

from idaes.core import FlowsheetBlock, LiquidPhase, Solute
from idaes.generic_models.unit_models import CSTR, PFR
from idaes.generic_models.properties.interrogator import (
        PropertyInterrogatorBlock, ReactionInterrogatorBlock)


@pytest.mark.unit
def test_interrogator_parameter_block():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(default={"dynamic": False})

    m.fs.params = PropertyInterrogatorBlock()
    m.fs.rxn_params = ReactionInterrogatorBlock(
            default={"property_package": m.fs.params})

    # Check that parameter block has expected attributes
    assert isinstance(m.fs.rxn_params.required_properties, dict)
    assert len(m.fs.rxn_params.required_properties) == 0


@pytest.mark.unit
def test_interrogator_rxn_block_unindexed_call():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(default={"dynamic": False})

    m.fs.params = PropertyInterrogatorBlock()
    m.fs.rxn_params = ReactionInterrogatorBlock(
            default={"property_package": m.fs.params})

    m.fs.props = m.fs.params.build_state_block([0])
    m.fs.rxns = m.fs.rxn_params.build_reaction_block(
            [0], default={"state_block": m.fs.props})

    # Check get_term methods return an unindexed dummy var
    assert m.fs.rxns[0].prop_unindexed is \
        m.fs.rxns[0]._dummy_var

    # Call again to make sure duplicates are skipped in required_properties
    assert m.fs.rxns[0].prop_unindexed is \
        m.fs.rxns[0]._dummy_var

    # Check that get_term calls were logged correctly
    assert m.fs.rxn_params.required_properties == {
            "prop_unindexed": ["fs.rxns"]}


@pytest.mark.unit
def test_interrogator_rxn_block_phase_call():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(default={"dynamic": False})

    m.fs.params = PropertyInterrogatorBlock()
    m.fs.rxn_params = ReactionInterrogatorBlock(
            default={"property_package": m.fs.params})

    m.fs.props = m.fs.params.build_state_block([0])
    m.fs.rxns = m.fs.rxn_params.build_reaction_block(
            [0], default={"state_block": m.fs.props})

    # Check get_term methods return an unindexed dummy var
    assert m.fs.rxns[0].prop_phase["Liq"] is \
        m.fs.rxns[0]._dummy_var_phase["Liq"]
    assert m.fs.rxns[0].prop_phase["Vap"] is \
        m.fs.rxns[0]._dummy_var_phase["Vap"]

    # Check that get_term calls were logged correctly
    assert m.fs.rxn_params.required_properties == {
            "prop_phase": ["fs.rxns"]}


@pytest.mark.unit
def test_interrogator_rxn_block_comp_call():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(default={"dynamic": False})

    m.fs.params = PropertyInterrogatorBlock()
    m.fs.rxn_params = ReactionInterrogatorBlock(
            default={"property_package": m.fs.params})

    m.fs.props = m.fs.params.build_state_block([0])
    m.fs.rxns = m.fs.rxn_params.build_reaction_block(
            [0], default={"state_block": m.fs.props})

    # Check get_term methods return an unindexed dummy var
    assert m.fs.rxns[0].prop_comp["A"] is \
        m.fs.rxns[0]._dummy_var_comp["A"]
    assert m.fs.rxns[0].prop_comp["B"] is \
        m.fs.rxns[0]._dummy_var_comp["B"]

    # Check that get_term calls were logged correctly
    assert m.fs.rxn_params.required_properties == {
            "prop_comp": ["fs.rxns"]}


@pytest.mark.unit
def test_interrogator_rxn_block_phase_comp_call():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(default={"dynamic": False})

    m.fs.params = PropertyInterrogatorBlock()
    m.fs.rxn_params = ReactionInterrogatorBlock(
            default={"property_package": m.fs.params})

    m.fs.props = m.fs.params.build_state_block([0])
    m.fs.rxns = m.fs.rxn_params.build_reaction_block(
            [0], default={"state_block": m.fs.props})

    # Check get_term methods return an unindexed dummy var
    assert m.fs.rxns[0].prop_phase_comp["Liq", "A"] is \
        m.fs.rxns[0]._dummy_var_phase_comp["Liq", "A"]
    assert m.fs.rxns[0].prop_phase_comp["Vap", "B"] is \
        m.fs.rxns[0]._dummy_var_phase_comp["Vap", "B"]

    # Check that get_term calls were logged correctly
    assert m.fs.rxn_params.required_properties == {
            "prop_phase_comp": ["fs.rxns"]}


@pytest.mark.unit
def test_interrogator_rxn_block_reaction_rate_call():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(default={"dynamic": False})

    m.fs.params = PropertyInterrogatorBlock()
    m.fs.rxn_params = ReactionInterrogatorBlock(
            default={"property_package": m.fs.params})

    m.fs.props = m.fs.params.build_state_block([0])
    m.fs.rxns = m.fs.rxn_params.build_reaction_block(
            [0], default={"state_block": m.fs.props})

    # Check get_term methods return an unindexed dummy var
    assert m.fs.rxns[0].reaction_rate["R1"] is \
        m.fs.rxns[0]._dummy_reaction_idx["R1"]
    assert m.fs.rxns[0].reaction_rate["R1"] is \
        m.fs.rxns[0]._dummy_reaction_idx["R1"]

    # Check that get_term calls were logged correctly
    assert m.fs.rxn_params.required_properties == {
            "reaction_rate": ["fs.rxns"]}


@pytest.mark.unit
def test_interrogator_rxn_block_dh_rxn_call():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(default={"dynamic": False})

    m.fs.params = PropertyInterrogatorBlock()
    m.fs.rxn_params = ReactionInterrogatorBlock(
            default={"property_package": m.fs.params})

    m.fs.props = m.fs.params.build_state_block([0])
    m.fs.rxns = m.fs.rxn_params.build_reaction_block(
            [0], default={"state_block": m.fs.props})

    # Check get_term methods return an unindexed dummy var
    assert m.fs.rxns[0].dh_rxn["R1"] is \
        m.fs.rxns[0]._dummy_reaction_idx["R1"]
    assert m.fs.rxns[0].dh_rxn["R1"] is \
        m.fs.rxns[0]._dummy_reaction_idx["R1"]

    # Check that get_term calls were logged correctly
    assert m.fs.rxn_params.required_properties == {
            "dh_rxn": ["fs.rxns"]}


@pytest.mark.unit
def test_interrogator_initialize_method():
    # Initialize method should return an TypeError
    m = ConcreteModel()
    m.fs = FlowsheetBlock(default={"dynamic": False})

    m.fs.params = PropertyInterrogatorBlock()
    m.fs.rxn_params = ReactionInterrogatorBlock(
            default={"property_package": m.fs.params})

    m.fs.props = m.fs.params.build_state_block([0])
    m.fs.rxns = m.fs.rxn_params.build_reaction_block(
            [0], default={"state_block": m.fs.props})

    with pytest.raises(TypeError,
                       match="Models constructed using the Reaction "
                       "Interrogator package cannot be used to solve a "
                       "flowsheet. Please rebuild your flowsheet using a "
                       "valid reaction package."):
        m.fs.rxns.initialize()


@pytest.fixture(scope="module")
def model():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(default={"dynamic": True,
                                   "time_units": pyunits.s})

    m.fs.params = PropertyInterrogatorBlock()
    m.fs.rxn_params = ReactionInterrogatorBlock(
            default={"property_package": m.fs.params})

    m.fs.R01 = CSTR(default={"property_package": m.fs.params,
                             "reaction_package": m.fs.rxn_params,
                             "has_heat_of_reaction": True})

    m.fs.R02 = PFR(default={"property_package": m.fs.params,
                            "reaction_package": m.fs.rxn_params})

    return m


# Test for physical parameters too, as these can not be checked without having
# a reaction package too.
@pytest.mark.component
def test_interrogate_flowsheet(model):
    assert model.fs.params.required_properties == {
            "material flow terms": ["fs.R01", "fs.R02"],
            "enthalpy flow terms": ["fs.R01", "fs.R02"],
            "material density terms": ["fs.R01", "fs.R02"],
            "energy density terms": ["fs.R01", "fs.R02"],
            "pressure": ["fs.R01", "fs.R02"]}

    assert model.fs.rxn_params.required_properties == {
            "reaction_rate": ["fs.R01", "fs.R02"],
            "dh_rxn": ["fs.R01"]}


@pytest.mark.component
def test_list_required_properties(model):
    prop_list = model.fs.params.list_required_properties()
    assert prop_list == ["material density terms",
                         "material flow terms",
                         "enthalpy flow terms",
                         "energy density terms",
                         "pressure"]

    rxn_list = model.fs.rxn_params.list_required_properties()
    assert rxn_list == ["dh_rxn", "reaction_rate"]


@pytest.mark.unit
def test_list_models_requiring_property(model):
    for k in model.fs.params.required_properties.keys():
        model_list = model.fs.params.list_models_requiring_property(k)
        assert model_list == ["fs.R01", "fs.R02"]


@pytest.mark.unit
def test_list_properties_required_by_model_by_name(model):
    prop_list = model.fs.rxn_params.list_properties_required_by_model("fs.R01")

    assert prop_list == ["dh_rxn",
                         "reaction_rate"]


@pytest.mark.unit
def test_list_properties_required_by_model_by_object(model):
    prop_list = model.fs.rxn_params.list_properties_required_by_model(
            model.fs.R01)

    assert prop_list == ["dh_rxn",
                         "reaction_rate"]


@pytest.mark.unit
def test_list_properties_required_by_model_invalid_model(model):
    with pytest.raises(ValueError):
        model.fs.rxn_params.list_properties_required_by_model("foo")


@pytest.mark.unit
def test_print_required_properties(model, capsys):
    model.fs.params.print_required_properties()
    model.fs.rxn_params.print_required_properties()

    captured = capsys.readouterr()
    assert captured.out == """
==========================================================================
Property Interrogator Summary

The Flowsheet requires the following properties (times required):

    material density terms                                               2
    material flow terms                                                  2
    enthalpy flow terms                                                  2
    energy density terms                                                 2
    pressure                                                             2

Note: User constraints may require additional properties which are not
reported here.

==========================================================================
Reaction Property Interrogator Summary

The Flowsheet requires the following reaction properties (times required):

    dh_rxn                                                               1
    reaction_rate                                                        2

Note: User constraints may require additional properties which are not
reported here.
"""


@pytest.mark.unit
def test_print_models_requiring_property(model, capsys):
    model.fs.rxn_params.print_models_requiring_property("reaction_rate")

    captured = capsys.readouterr()
    assert captured.out == """
The following models in the Flowsheet require reaction_rate:
    fs.R01
    fs.R02
"""


@pytest.mark.unit
def test_print_properties_reqruied_by_model(model, capsys):
    model.fs.rxn_params.print_properties_required_by_model("fs.R01")

    captured = capsys.readouterr()
    assert captured.out == """
The following reaction properties are required by model fs.R01:
    dh_rxn
    reaction_rate
"""


@pytest.mark.unit
def test_interrogator_rxn_block_unindexed_call_custom_phase_comp():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(default={"dynamic": False})

    m.fs.params = PropertyInterrogatorBlock(default={
        "phase_list": {"P1": LiquidPhase, "P2": None},
        "component_list": {"c1": Solute, "c2": None}})
    m.fs.rxn_params = ReactionInterrogatorBlock(
            default={"property_package": m.fs.params})

    m.fs.props = m.fs.params.build_state_block([0])
    m.fs.rxns = m.fs.rxn_params.build_reaction_block(
            [0], default={"state_block": m.fs.props})

    # Check phase and component lists
    assert m.fs.rxn_params.phase_list == ["P1", "P2"]
    assert m.fs.rxn_params.component_list == ["c1", "c2"]

    # Check get_term methods return an unindexed dummy var
    assert m.fs.rxns[0].prop_unindexed is \
        m.fs.rxns[0]._dummy_var

    # Call again to make sure duplicates are skipped in required_properties
    assert m.fs.rxns[0].prop_unindexed is \
        m.fs.rxns[0]._dummy_var

    # Check that get_term calls were logged correctly
    assert m.fs.rxn_params.required_properties == {
            "prop_unindexed": ["fs.rxns"]}



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES), and is copyright (c) 2018-2021
# by the software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia University
# Research Corporation, et al.  All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and
# license information.
#################################################################################
"""
Basic IDAES 1D Heat Exchanger Model.

1D Single pass shell and tube HX model with 0D wall conduction model
"""
# Import Python libraries
from enum import Enum

# Import Pyomo libraries
from pyomo.environ import (
    SolverFactory,
    Var,
    Constraint,
    value,
    units as pyunits
)
from pyomo.common.config import ConfigBlock, ConfigValue, In

# Import IDAES cores
from idaes.core import (
    ControlVolume1DBlock,
    UnitModelBlockData,
    declare_process_block_class,
    MaterialBalanceType,
    EnergyBalanceType,
    MomentumBalanceType,
    FlowDirection,
    useDefault,
)
from idaes.generic_models.unit_models.heat_exchanger \
    import HeatExchangerFlowPattern
from idaes.core.util.config import is_physical_parameter_block
from idaes.core.util.misc import add_object_reference
from idaes.core.util.exceptions import ConfigurationError
from idaes.core.util.tables import create_stream_table_dataframe
from idaes.core.util.constants import Constants as c
from idaes.core.util import get_solver, scaling as iscale

import idaes.logger as idaeslog


__author__ = "Jaffer Ghouse"

# Set up logger
_log = idaeslog.getLogger(__name__)


class WallConductionType(Enum):
    zero_dimensional = 0
    one_dimensional = 1
    two_dimensional = 2


@declare_process_block_class("HeatExchanger1D")
class HeatExchanger1DData(UnitModelBlockData):
    """Standard Heat Exchanger 1D Unit Model Class."""

    CONFIG = UnitModelBlockData.CONFIG()
    # Template for config arguments for shell and tube side
    _SideTemplate = ConfigBlock()
    _SideTemplate.declare(
        "dynamic",
        ConfigValue(
            default=useDefault,
            domain=In([useDefault, True, False]),
            description="Dynamic model flag",
            doc="""Indicates whether this model will be dynamic or not,
**default** = useDefault.
**Valid values:** {
**useDefault** - get flag from parent (default = False),
**True** - set as a dynamic model,
**False** - set as a steady-state model.}""",
        ),
    )
    _SideTemplate.declare(
        "has_holdup",
        ConfigValue(
            default=useDefault,
            domain=In([useDefault, True, False]),
            description="Holdup construction flag",
            doc="""Indicates whether holdup terms should be constructed or not.
Must be True if dynamic = True,
**default** - False.
**Valid values:** {
**useDefault** - get flag from parent (default = False),
**True** - construct holdup terms,
**False** - do not construct holdup terms}""",
        ),
    )
    _SideTemplate.declare(
        "material_balance_type",
        ConfigValue(
            default=MaterialBalanceType.useDefault,
            domain=In(MaterialBalanceType),
            description="Material balance construction flag",
            doc="""Indicates what type of mass balance should be constructed,
**default** - MaterialBalanceType.useDefault.
**Valid values:** {
**MaterialBalanceType.useDefault - refer to property package for default
balance type
**MaterialBalanceType.none** - exclude material balances,
**MaterialBalanceType.componentPhase** - use phase component balances,
**MaterialBalanceType.componentTotal** - use total component balances,
**MaterialBalanceType.elementTotal** - use total element balances,
**MaterialBalanceType.total** - use total material balance.}""",
        ),
    )
    _SideTemplate.declare(
        "energy_balance_type",
        ConfigValue(
            default=EnergyBalanceType.useDefault,
            domain=In(EnergyBalanceType),
            description="Energy balance construction flag",
            doc="""Indicates what type of energy balance should be constructed,
**default** - EnergyBalanceType.useDefault.
**Valid values:** {
**EnergyBalanceType.useDefault - refer to property package for default
balance type
**EnergyBalanceType.none** - exclude energy balances,
**EnergyBalanceType.enthalpyTotal** - single enthalpy balance for material,
**EnergyBalanceType.enthalpyPhase** - enthalpy balances for each phase,
**EnergyBalanceType.energyTotal** - single energy balance for material,
**EnergyBalanceType.energyPhase** - energy balances for each phase.}""",
        ),
    )
    _SideTemplate.declare(
        "momentum_balance_type",
        ConfigValue(
            default=MomentumBalanceType.pressureTotal,
            domain=In(MomentumBalanceType),
            description="Momentum balance construction flag",
            doc="""Indicates what type of momentum balance should
be constructed,
**default** - MomentumBalanceType.pressureTotal.
**Valid values:** {
**MomentumBalanceType.none** - exclude momentum balances,
**MomentumBalanceType.pressureTotal** - single pressure balance for material,
**MomentumBalanceType.pressurePhase** - pressure balances for each phase,
**MomentumBalanceType.momentumTotal** - single momentum balance for material,
**MomentumBalanceType.momentumPhase** - momentum balances for each phase.}""",
        ),
    )
    _SideTemplate.declare(
        "has_pressure_change",
        ConfigValue(
            default=False,
            domain=In([True, False]),
            description="Pressure change term construction flag",
            doc="""Indicates whether terms for pressure change should be
constructed,
**default** - False.
**Valid values:** {
**True** - include pressure change terms,
**False** - exclude pressure change terms.}""",
        ),
    )
    _SideTemplate.declare(
        "has_phase_equilibrium",
        ConfigValue(
            default=False,
            domain=In([True, False]),
            description="Phase equilibrium term construction flag",
            doc="""Argument to enable phase equilibrium on the shell side.
- True - include phase equilibrium term
- False - do not include phase equilibrium term""",
        ),
    )
    _SideTemplate.declare(
        "property_package",
        ConfigValue(
            default=None,
            domain=is_physical_parameter_block,
            description="Property package to use for control volume",
            doc="""Property parameter object used to define property
calculations
(default = 'use_parent_value')
- 'use_parent_value' - get package from parent (default = None)
- a ParameterBlock object""",
        ),
    )
    _SideTemplate.declare(
        "property_package_args",
        ConfigValue(
            default={},
            description="Arguments for constructing shell property package",
            doc="""A dict of arguments to be passed to the PropertyBlockData
and used when constructing these
(default = 'use_parent_value')
- 'use_parent_value' - get package from parent (default = None)
- a dict (see property package for documentation)""",
        ),
    )
    # TODO : We should probably think about adding a consistency check for the
    # TODO : discretisation methods as well.
    _SideTemplate.declare(
        "transformation_method",
        ConfigValue(
            default=useDefault,
            description="Discretization method to use for DAE transformation",
            doc="""Discretization method to use for DAE transformation. See
Pyomo documentation for supported transformations.""",
        ),
    )
    _SideTemplate.declare(
        "transformation_scheme",
        ConfigValue(
            default=useDefault,
            description="Discretization scheme to use for DAE transformation",
            doc="""Discretization scheme to use when transformating domain. See
Pyomo documentation for supported schemes.""",
        ),
    )

    # Create individual config blocks for shell and tube side
    CONFIG.declare(
        "shell_side", _SideTemplate(doc="shell side config arguments"))
    CONFIG.declare(
        "tube_side", _SideTemplate(doc="tube side config arguments"))

    # Common config args for both sides
    CONFIG.declare(
        "finite_elements",
        ConfigValue(
            default=20,
            domain=int,
            description="Number of finite elements length domain",
            doc="""Number of finite elements to use when discretizing length
domain (default=20)""",
        ),
    )
    CONFIG.declare(
        "collocation_points",
        ConfigValue(
            default=5,
            domain=int,
            description="Number of collocation points per finite element",
            doc="""Number of collocation points to use per finite element when
discretizing length domain (default=3)""",
        ),
    )
    CONFIG.declare(
        "flow_type",
        ConfigValue(
            default=HeatExchangerFlowPattern.cocurrent,
            domain=In(HeatExchangerFlowPattern),
            description="Flow configuration of heat exchanger",
            doc="""Flow configuration of heat exchanger
- HeatExchangerFlowPattern.cocurrent: shell and tube flows from 0 to 1
(default)
- HeatExchangerFlowPattern.countercurrent: shell side flows from 0 to 1
tube side flows from 1 to 0""",
        ),
    )
    CONFIG.declare(
        "has_wall_conduction",
        ConfigValue(
            default=WallConductionType.zero_dimensional,
            domain=In(WallConductionType),
            description="Conduction model for tube wall",
            doc="""Argument to enable type of wall heat conduction model.
- WallConductionType.zero_dimensional - 0D wall model (default),
- WallConductionType.one_dimensional - 1D wall model along the thickness of the
tube,
- WallConductionType.two_dimensional - 2D wall model along the lenghth and
thickness of the tube""",
        ),
    )

    def build(self):
        """
        Begin building model (pre-DAE transformation).

        Args:
            None

        Returns:
            None
        """
        # Call UnitModel.build to setup dynamics
        super(HeatExchanger1DData, self).build()

        # Set flow directions for the control volume blocks and specify
        # dicretisation if not specified.
        if self.config.flow_type == HeatExchangerFlowPattern.cocurrent:
            set_direction_shell = FlowDirection.forward
            set_direction_tube = FlowDirection.forward
            if (
                self.config.shell_side.transformation_method
                != self.config.tube_side.transformation_method
            ) or (
                self.config.shell_side.transformation_scheme
                != self.config.tube_side.transformation_scheme
            ):
                raise ConfigurationError(
                    "HeatExchanger1D only supports similar transformation "
                    "schemes on the shell side and tube side domains for "
                    "both cocurrent and countercurrent flow patterns."
                )
            if self.config.shell_side.transformation_method is useDefault:
                _log.warning(
                    "Discretization method was "
                    "not specified for the shell side of the "
                    "co-current heat exchanger. "
                    "Defaulting to finite "
                    "difference method on the shell side."
                )
                self.config.shell_side.transformation_method = \
                    "dae.finite_difference"
            if self.config.tube_side.transformation_method is useDefault:
                _log.warning(
                    "Discretization method was "
                    "not specified for the tube side of the "
                    "co-current heat exchanger. "
                    "Defaulting to finite "
                    "difference method on the tube side."
                )
                self.config.tube_side.transformation_method = \
                    "dae.finite_difference"
            if self.config.shell_side.transformation_scheme is useDefault:
                _log.warning(
                    "Discretization scheme was "
                    "not specified for the shell side of the "
                    "co-current heat exchanger. "
                    "Defaulting to backward finite "
                    "difference on the shell side."
                )
                self.config.shell_side.transformation_scheme = "BACKWARD"
            if self.config.tube_side.transformation_scheme is useDefault:
                _log.warning(
                    "Discretization scheme was "
                    "not specified for the tube side of the "
                    "co-current heat exchanger. "
                    "Defaulting to backward finite "
                    "difference on the tube side."
                )
                self.config.tube_side.transformation_scheme = "BACKWARD"
        elif self.config.flow_type == HeatExchangerFlowPattern.countercurrent:
            set_direction_shell = FlowDirection.forward
            set_direction_tube = FlowDirection.backward
            if self.config.shell_side.transformation_method is useDefault:
                _log.warning(
                    "Discretization method was "
                    "not specified for the shell side of the "
                    "counter-current heat exchanger. "
                    "Defaulting to finite "
                    "difference method on the shell side."
                )
                self.config.shell_side.transformation_method = \
                    "dae.finite_difference"
            if self.config.tube_side.transformation_method is useDefault:
                _log.warning(
                    "Discretization method was "
                    "not specified for the tube side of the "
                    "counter-current heat exchanger. "
                    "Defaulting to finite "
                    "difference method on the tube side."
                )
                self.config.tube_side.transformation_method = \
                    "dae.finite_difference"
            if self.config.shell_side.transformation_scheme is useDefault:
                _log.warning(
                    "Discretization scheme was "
                    "not specified for the shell side of the "
                    "counter-current heat exchanger. "
                    "Defaulting to backward finite "
                    "difference on the shell side."
                )
                self.config.shell_side.transformation_scheme = "BACKWARD"
            if self.config.tube_side.transformation_scheme is useDefault:
                _log.warning(
                    "Discretization scheme was "
                    "not specified for the tube side of the "
                    "counter-current heat exchanger. "
                    "Defaulting to forward finite "
                    "difference on the tube side."
                )
                self.config.tube_side.transformation_scheme = "BACKWARD"
        else:
            raise ConfigurationError(
                "{} HeatExchanger1D only supports cocurrent and "
                "countercurrent flow patterns, but flow_type configuration"
                " argument was set to {}."
                .format(self.name, self.config.flow_type)
            )

        # Control volume 1D for shell
        self.shell = ControlVolume1DBlock(
            default={
                "dynamic": self.config.shell_side.dynamic,
                "has_holdup": self.config.shell_side.has_holdup,
                "property_package": self.config.shell_side.property_package,
                "property_package_args":
                self.config.shell_side.property_package_args,
                "transformation_method":
                self.config.shell_side.transformation_method,
                "transformation_scheme":
                self.config.shell_side.transformation_scheme,
                "finite_elements": self.config.finite_elements,
                "collocation_points": self.config.collocation_points,
            }
        )

        self.tube = ControlVolume1DBlock(
            default={
                "dynamic": self.config.tube_side.dynamic,
                "has_holdup": self.config.tube_side.has_holdup,
                "property_package": self.config.tube_side.property_package,
                "property_package_args":
                self.config.tube_side.property_package_args,
                "transformation_method":
                self.config.tube_side.transformation_method,
                "transformation_scheme":
                self.config.tube_side.transformation_scheme,
                "finite_elements": self.config.finite_elements,
                "collocation_points": self.config.collocation_points,
            }
        )

        self.shell.add_geometry(flow_direction=set_direction_shell)
        self.tube.add_geometry(flow_direction=set_direction_tube)

        self.shell.add_state_blocks(
            information_flow=set_direction_shell,
            has_phase_equilibrium=self.config.shell_side.has_phase_equilibrium,
        )
        self.tube.add_state_blocks(
            information_flow=set_direction_tube,
            has_phase_equilibrium=self.config.tube_side.has_phase_equilibrium,
        )

        # Populate shell
        self.shell.add_material_balances(
            balance_type=self.config.shell_side.material_balance_type,
            has_phase_equilibrium=self.config.shell_side.has_phase_equilibrium,
        )

        self.shell.add_energy_balances(
            balance_type=self.config.shell_side.energy_balance_type,
            has_heat_transfer=True,
        )

        self.shell.add_momentum_balances(
            balance_type=self.config.shell_side.momentum_balance_type,
            has_pressure_change=self.config.shell_side.has_pressure_change,
        )

        self.shell.apply_transformation()

        # Populate tube
        self.tube.add_material_balances(
            balance_type=self.config.tube_side.material_balance_type,
            has_phase_equilibrium=self.config.tube_side.has_phase_equilibrium,
        )

        self.tube.add_energy_balances(
            balance_type=self.config.tube_side.energy_balance_type,
            has_heat_transfer=True,
        )

        self.tube.add_momentum_balances(
            balance_type=self.config.tube_side.momentum_balance_type,
            has_pressure_change=self.config.tube_side.has_pressure_change,
        )

        self.tube.apply_transformation()

        # Add Ports for shell side
        self.add_inlet_port(name="shell_inlet", block=self.shell)
        self.add_outlet_port(name="shell_outlet", block=self.shell)

        # Add Ports for tube side
        self.add_inlet_port(name="tube_inlet", block=self.tube)
        self.add_outlet_port(name="tube_outlet", block=self.tube)

        # Add reference to control volume geometry
        add_object_reference(self, "shell_area", self.shell.area)
        add_object_reference(self, "shell_length", self.shell.length)
        add_object_reference(self, "tube_area", self.tube.area)
        add_object_reference(self, "tube_length", self.tube.length)

        self._make_performance()

    def _make_performance(self):
        """
        Constraints for unit model.

        Args:
            None

        Returns:
            None
        """
        shell_units = self.config.shell_side.property_package.\
            get_metadata().get_derived_units
        tube_units = self.config.tube_side.property_package.\
            get_metadata().get_derived_units

        # Unit model variables
        # HX dimensions
        self.d_shell = Var(initialize=1,
                           doc="Diameter of shell",
                           units=shell_units("length"))
        self.d_tube_outer = Var(initialize=0.011,
                                doc="Outer diameter of tube",
                           units=shell_units("length"))
        self.d_tube_inner = Var(initialize=0.010,
                                doc="Inner diameter of tube",
                           units=shell_units("length"))
        self.N_tubes = Var(initialize=1,
                           doc="Number of tubes",
                           units=pyunits.dimensionless)

        # Note: In addition to the above variables, "shell_length" and
        # "tube_length" need to be fixed at the flowsheet level

        # Performance variables
        self.shell_heat_transfer_coefficient = Var(
            self.flowsheet().time,
            self.shell.length_domain,
            initialize=50,
            doc="Heat transfer coefficient",
            units=shell_units("heat_transfer_coefficient")
        )
        self.tube_heat_transfer_coefficient = Var(
            self.flowsheet().time,
            self.tube.length_domain,
            initialize=50,
            doc="Heat transfer coefficient",
            units=tube_units("heat_transfer_coefficient")
        )

        # Wall 0D model (Q_shell = Q_tube*N_tubes)
        if self.config.has_wall_conduction == \
                WallConductionType.zero_dimensional:
            self.temperature_wall = Var(
                self.flowsheet().time,
                self.tube.length_domain,
                initialize=298.15,
                units=shell_units("temperature")
            )

            # Performance equations
            # Energy transfer between shell and tube wall

            @self.Constraint(
                self.flowsheet().time,
                self.shell.length_domain,
                doc="Heat transfer between shell and tube",
            )
            def shell_heat_transfer_eq(self, t, x):
                return self.shell.heat[t, x] == -self.N_tubes * (
                    self.shell_heat_transfer_coefficient[t, x]
                    * c.pi
                    * self.d_tube_outer
                    * (
                        self.shell.properties[t, x].temperature
                        - self.temperature_wall[t, x]
                    )
                )

            # Energy transfer between tube wall and tube
            @self.Constraint(
                self.flowsheet().time,
                self.tube.length_domain,
                doc="Convective heat transfer",
            )
            def tube_heat_transfer_eq(self, t, x):
                return self.tube.heat[t, x] == \
                    self.tube_heat_transfer_coefficient[
                    t, x
                ] * c.pi * pyunits.convert(self.d_tube_inner,
                                           to_units=tube_units("length")) * (
                    pyunits.convert(self.temperature_wall[t, x],
                                    to_units=tube_units('temperature')) -
                    self.tube.properties[t, x].temperature
                )

            if shell_units("length") is None:
                # Backwards compatability check
                q_units = None
            else:
                q_units = shell_units("power")/shell_units("length")
            # Wall 0D model
            @self.Constraint(
                self.flowsheet().time,
                self.shell.length_domain,
                doc="wall 0D model",
            )
            def wall_0D_model(self, t, x):
                return pyunits.convert(self.tube.heat[t, x],
                                       to_units=q_units) == -(
                            self.shell.heat[t, x] / self.N_tubes)

        else:
            raise NotImplementedError(
                "{} HeatExchanger1D has not yet implemented support for "
                "wall conduction models."
            )

        # Define tube area in terms of tube diameter
        self.area_calc_tube = Constraint(
            expr=4 * self.tube_area == c.pi * pyunits.convert(
                self.d_tube_inner, to_units=tube_units("length"))**2
        )

        # Define shell area in terms of shell and tube diameter
        self.area_calc_shell = Constraint(
            expr=4 * self.shell_area
            == c.pi * (self.d_shell**2 - self.N_tubes*self.d_tube_outer**2)
        )

    def initialize(
        self,
        shell_state_args=None,
        tube_state_args=None,
        outlvl=idaeslog.NOTSET,
        solver=None,
        optarg=None,
    ):
        """
        Initialization routine for the unit.

        Keyword Arguments:
            state_args : a dict of arguments to be passed to the property
                         package(s) to provide an initial state for
                         initialization (see documentation of the specific
                         property package) (default = {}).
            outlvl : sets output level of initialization routine
            optarg : solver options dictionary object (default=None, use
                     default solver options)
            solver : str indicating which solver to use during
                     initialization (default = None, use default solver)

        Returns:
            None
        """
        init_log = idaeslog.getInitLogger(self.name, outlvl, tag="unit")
        solve_log = idaeslog.getSolveLogger(self.name, outlvl, tag="unit")

        # Create solver
        opt = get_solver(solver, optarg)

        # ---------------------------------------------------------------------
        # Initialize shell block
        flags_shell = self.shell.initialize(
            outlvl=outlvl,
            optarg=optarg,
            solver=solver,
            state_args=shell_state_args,
        )

        flags_tube = self.tube.initialize(
            outlvl=outlvl,
            optarg=optarg,
            solver=solver,
            state_args=tube_state_args,
        )

        init_log.info_high("Initialization Step 1 Complete.")

        # ---------------------------------------------------------------------
        # Solve unit
        # Wall 0D
        if self.config.has_wall_conduction == \
            WallConductionType.zero_dimensional:
            shell_units = self.config.shell_side.property_package.\
                get_metadata().get_derived_units
            for t in self.flowsheet().time:
                for z in self.shell.length_domain:
                    self.temperature_wall[t, z].fix(
                        value(
                            0.5
                            * (
                                self.shell.properties[t, 0].temperature
                                + pyunits.convert(
                                    self.tube.properties[t, 0].temperature,
                                    to_units=shell_units('temperature'))
                            )
                        )
                    )

            self.tube.deactivate()
            self.tube_heat_transfer_eq.deactivate()
            self.wall_0D_model.deactivate()

            with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
                res = opt.solve(self, tee=slc.tee)
            init_log.info_high(
                "Initialization Step 2 {}.".format(idaeslog.condition(res))
            )

            self.tube.activate()
            self.tube_heat_transfer_eq.activate()

            with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
                res = opt.solve(self, tee=slc.tee)
            init_log.info_high(
                "Initialization Step 3 {}.".format(idaeslog.condition(res))
            )

            self.wall_0D_model.activate()
            self.temperature_wall.unfix()

            with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
                res = opt.solve(self, tee=slc.tee)
            init_log.info_high(
                "Initialization Step 4 {}.".format(idaeslog.condition(res))
            )

        self.shell.release_state(flags_shell)
        self.tube.release_state(flags_tube)

        init_log.info("Initialization Complete.")

    def _get_performance_contents(self, time_point=0):
        var_dict = {}
        var_dict["Shell Area"] = self.shell.area
        var_dict["Shell Diameter"] = self.d_shell
        var_dict["Shell Length"] = self.shell.length
        var_dict["Tube Area"] = self.tube.area
        var_dict["Tube Outer Diameter"] = self.d_tube_outer
        var_dict["Tube Inner Diameter"] = self.d_tube_inner
        var_dict["Tube Length"] = self.tube.length
        var_dict["Number of Tubes"] = self.N_tubes

        return {"vars": var_dict}

    def _get_stream_table_contents(self, time_point=0):
        return create_stream_table_dataframe(
            {
                "Shell Inlet": self.shell_inlet,
                "Shell Outlet": self.shell_outlet,
                "Tube Inlet": self.tube_inlet,
                "Tube Outlet": self.tube_outlet,
            },
            time_point=time_point,
        )

    def calculate_scaling_factors(self):
        super().calculate_scaling_factors()

        for i, c in self.shell_heat_transfer_eq.items():
            iscale.constraint_scaling_transform(
                c, iscale.get_scaling_factor(
                    self.shell.heat[i], default=1, warning=True),
                overwrite=False)

        for i, c in self.tube_heat_transfer_eq.items():
            iscale.constraint_scaling_transform(c, iscale.get_scaling_factor(
                    self.tube.heat[i], default=1, warning=True),
                overwrite=False)



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2020, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
Case B31A - Natural Gas Combined Cycle (NGCC) plant_gross_power
Reference: NETL-PUB-22638
Cost and Performance Baseline for Fossil Energy Plants Volume 1:
Bituminous Coal and Natural Gas to Electricity; https://doi.org/10.2172/1569246
Author: A. Deshpande and M. Zamarripa
"""
import pytest
from idaes.power_generation.costing.power_plant_costing import \
     (get_PP_costing,
      build_flowsheet_cost_constraint,
      costing_initialization)
from idaes.core.util.model_statistics import degrees_of_freedom
import pyomo.environ as pyo
from idaes.core import FlowsheetBlock
from idaes.core.util import get_solver

# Get default solver for testing
solver = get_solver()


@pytest.fixture(scope="module")
def build_costing():
    # Create a Concrete Model as the top level object
    m = pyo.ConcreteModel()

    # Add a flowsheet object to the model
    m.fs = FlowsheetBlock(default={"dynamic": False})
    m.fs.get_costing(year='2018')

    return m


@pytest.mark.unit
def test_get_costing(build_costing):
    m = build_costing
    assert hasattr(m.fs.costing, "CE_index")
    # Accounts with Feedwater Flow to HP section of HRSG, as the
    # reference/scaling parameter - Exhibit 5-15
    FW_accounts = ['3.1', '3.3', '8.4']

    m.fs.b1 = pyo.Block()
    m.fs.b1.feedwater_flowrate = pyo.Var(initialize=1085751)  # lb/hr
    m.fs.b1.feedwater_flowrate.fix()
    get_PP_costing(m.fs.b1, FW_accounts,
                   m.fs.b1.feedwater_flowrate, 'lb/hr', 6)
    assert isinstance(m.fs.b1.costing.total_plant_cost, pyo.Var)
    assert hasattr(m.fs.b1.costing, "bare_erected_cost")
    assert isinstance(m.fs.b1.costing.total_plant_cost_eq, pyo.Constraint)
    assert degrees_of_freedom(m) == 0


@pytest.mark.skipif(solver is None, reason="Solver not available")
@pytest.mark.component
def test_units1_costing(build_costing):
    m = build_costing

    # Accounts with Feedwater Flow to HP section of HRSG, as the
    # reference/scaling parameter - Exhibit 5-15
    FW_accounts = ['3.1', '3.3', '8.4']

    # Accounts with Raw water withdrawal as the reference/scaling parameter
    # Exhibit 5-14
    RW_withdraw_accounts = ['3.2', '3.4', '3.5', '9.5', '14.6']
    m.fs.b2 = pyo.Block()

    m.fs.b2.raw_water_withdrawal = pyo.Var(initialize=2902)  # gpm
    m.fs.b2.raw_water_withdrawal.fix()
    get_PP_costing(m.fs.b2, RW_withdraw_accounts,
                   m.fs.b2.raw_water_withdrawal, 'gpm', 6)

    # Accounts with fuel gas flowrate as the reference/scaling parameter
    # Exhibit 5-15 stream 2, Exhibit 5-8
    FuelG_accounts = ['3.6', '3.9', '6.1', '6.3', '6.4']
    m.fs.b3 = pyo.Block()
    # Obtain Fuel gas flowrate in acm
    fuelgas_value = 205630  # lb/hr

    m.fs.b3.fg_flowrate = pyo.Var(initialize=fuelgas_value)  # lb/hr
    m.fs.b3.fg_flowrate.fix()
    get_PP_costing(m.fs.b3, FuelG_accounts,
                   m.fs.b3.fg_flowrate, 'lb/hr', 6)

    # Accounts with process water discharge as the reference/scaling parameter
    # Exhibit 5-14
    PW_discharge_accounts = ['3.7']
    m.fs.b4 = pyo.Block()

    m.fs.b4.process_water_discharge = pyo.Var(initialize=657)  # gpm
    m.fs.b4.process_water_discharge.fix()
    get_PP_costing(m.fs.b4, PW_discharge_accounts,
                   m.fs.b4.process_water_discharge, 'gpm', 6)

    # Initialize costing
    costing_initialization(m.fs)
    assert degrees_of_freedom(m) == 0

    # Solve the model
    results = solver.solve(m, tee=True)
    assert results.solver.termination_condition == \
        pyo.TerminationCondition.optimal
    assert results.solver.status == pyo.SolverStatus.ok

     # Accounts with raw water withdrawal as reference parameter
    assert pytest.approx(26.435, abs=0.5) \
        == sum(pyo.value(m.fs.b2.costing.total_plant_cost[ac])
               for ac in RW_withdraw_accounts)
    # Accounts with fuel gas as reference parameter
    assert pytest.approx(158.415, abs=0.5) \
        == sum(pyo.value(m.fs.b3.costing.total_plant_cost[ac])
               for ac in FuelG_accounts)

    # Accounts with process water discharge as reference parameter
    assert pytest.approx(11.608, abs=0.5) \
        == sum(pyo.value(m.fs.b4.costing.total_plant_cost[ac])
               for ac in PW_discharge_accounts)


@pytest.mark.skipif(solver is None, reason="Solver not available")
@pytest.mark.component
def test_units2_costing(build_costing):
    m = build_costing
    # Accounts with flue gas flowrate as the reference/scaling parameter
    # Exhibit 5-15 stream 3, Exhibit 5-8
    FG_accounts = ['7.6']
    m.fs.b5 = pyo.Block()
    # Obtain Flue gas flowrate in acm
    fluegas_value = (8658430/60)/0.025  # ft3/min

    m.fs.b5.fg_flowrate = pyo.Var(initialize=fluegas_value)  # ft3/min
    m.fs.b5.fg_flowrate.fix()
    get_PP_costing(m.fs.b5, FG_accounts,
                   m.fs.b5.fg_flowrate, 'acfm', 6)

    # Accounts with combustion turbine gross power as the reference/scaling
    # parameter
    # Exhibit 5-9
    CT_grosspower_accounts = ['6.5']
    m.fs.b6 = pyo.Block()
    # Obtain combustion turbine gross power in kW
    CT_gross_power = 477*1000  # kW

    m.fs.b6.ct_gross_power = pyo.Var(initialize=CT_gross_power)  # kW
    m.fs.b6.ct_gross_power.fix()
    get_PP_costing(m.fs.b6, CT_grosspower_accounts,
                   m.fs.b6.ct_gross_power, 'kW', 6)

    # Accounts with HRSG duty as the reference/scaling parameter
    # Exhibit 5-8, streams 3 and 4
    HRSG_duty_accounts = ['7.1', '7.2']
    m.fs.b7 = pyo.Block()
    # Obtain HRSG duty in MMBtu/hr, overall energy balance
    HRSG_duty = -(-538.1+277.1)*8658430/(10**6)  # MMBtu/hr

    m.fs.b7.hrsg_duty = pyo.Var(initialize=HRSG_duty)  # MMBtu/hr
    m.fs.b7.hrsg_duty.fix()
    get_PP_costing(m.fs.b7, HRSG_duty_accounts,
                   m.fs.b7.hrsg_duty, 'MMBtu/hr', 6)

    # Accounts with gas flow to stack as the reference/scaling parameter
    # Exhibit 5-8, stream 4
    Stack_flow_gas_accounts = ['7.3', '7.4', '7.5']
    m.fs.b8 = pyo.Block()
    # Obtain gas flowrate to stack in ft3/min
    stack_flow_gas = (8658430/60)/0.061  # ft3/min

    m.fs.b8.stack_flow_gas = pyo.Var(initialize=stack_flow_gas)  # ft3/min
    m.fs.b8.stack_flow_gas.fix()
    get_PP_costing(m.fs.b8, Stack_flow_gas_accounts,
                   m.fs.b8.stack_flow_gas, 'acfm', 6)

    # Initialize costing
    costing_initialization(m.fs)
    assert degrees_of_freedom(m) == 0

    # Solve the model
    results = solver.solve(m, tee=True)
    assert results.solver.termination_condition == \
        pyo.TerminationCondition.optimal
    assert results.solver.status == pyo.SolverStatus.ok
    # Accounts with HRSG duty as reference parameter
    assert pytest.approx(90.794, abs=0.1) \
        == sum(pyo.value(m.fs.b7.costing.total_plant_cost[ac])
               for ac in HRSG_duty_accounts)


@pytest.mark.skipif(solver is None, reason="Solver not available")
@pytest.mark.component
def test_units3_costing(build_costing):
    m = build_costing
    # Accounts with steam turbine gross power as the reference/scaling
    # parameter
    # Exhibit 5-9
    Steam_turbine_gross_power_accounts = ['8.1', '8.2', '8.5', '14.3']
    m.fs.b9 = pyo.Block()
    # Obtain steam turbine gross power in kW
    ST_gross_power = 263*1000  # kW

    m.fs.b9.st_gross_power = pyo.Var(initialize=ST_gross_power)  # kW
    m.fs.b9.st_gross_power.fix()
    get_PP_costing(m.fs.b9, Steam_turbine_gross_power_accounts,
                   m.fs.b9.st_gross_power, 'kW', 6)

    # Accounts with condenser duty as the reference/scaling parameter
    # Exhibit 5-9
    Condenser_duty_accounts = ['8.3']
    m.fs.b10 = pyo.Block()
    # Obtain condenser duty in MMBtu/hr
    condenser_duty = 1332  # MMBtu/hr

    m.fs.b10.cond_duty = pyo.Var(initialize=condenser_duty)  # MMBtu/hr
    m.fs.b10.cond_duty.fix()
    get_PP_costing(m.fs.b10, Condenser_duty_accounts,
                   m.fs.b10.cond_duty, 'MMBtu/hr', 6)

    # Accounts with cooling tower duty as the reference/scaling parameter
    # Exhibit 5-16
    Cooling_tower_accounts = ['9.1']
    m.fs.b11 = pyo.Block()
    # Obtain cooling tower duty in MMBtu/hr (includes condenser, Acid gas
    # removal, and other cooling loads)
    cooling_tower_duty = 1357  # MMBtu/hr

    m.fs.b11.cool_tower_duty = pyo.Var(initialize=cooling_tower_duty)
    # MMBtu/hr
    m.fs.b11.cool_tower_duty.fix()
    get_PP_costing(m.fs.b11, Cooling_tower_accounts,
                   m.fs.b11.cool_tower_duty, 'MMBtu/hr', 6)

    # Accounts with circulating water flowrate as the reference/scaling
    # parameter
    Circ_water_accounts = ['9.2', '9.3', '9.4', '9.6', '9.7', '14.5']
    m.fs.b12 = pyo.Block()
    # Obtain circulating water flowrate in gpm
    cir_water_flowrate = 217555  # gpm

    m.fs.b12.circ_water_flow = pyo.Var(initialize=cir_water_flowrate)  # gpm
    m.fs.b12.circ_water_flow.fix()
    get_PP_costing(m.fs.b12, Circ_water_accounts,
                   m.fs.b12.circ_water_flow, 'gpm', 6)

    # Accounts with total plant gross power as the reference/scaling parameter
    # Exhibit 5-9
    plant_gross_power_accounts = ['11.1', '11.7', '11.9', '13.1', '13.2',
                                  '13.3', '14.4', '14.7', '14.8', '14.9',
                                  '14.10']
    m.fs.b13 = pyo.Block()
    # Obtain total plant gross power in kW
    plant_gross_power = 740000  # kW

    m.fs.b13.gross_power = pyo.Var(initialize=plant_gross_power)  # kW
    m.fs.b13.gross_power.fix()
    get_PP_costing(m.fs.b13, plant_gross_power_accounts,
                   m.fs.b13.gross_power, 'kW', 6)

    # Accounts with auxilliary load as the reference/scaling parameter
    # Exhibit 5-9
    auxilliary_load_accounts = ['11.2', '11.3', '11.4', '11.5', '11.6', '12.1',
                                '12.2', '12.3', '12.4', '12.5', '12.6', '12.7',
                                '12.8', '12.9']
    m.fs.b14 = pyo.Block()
    # Obtain auxilliary load in kW
    aux_load = 14*1000  # kW

    m.fs.b14.auxilliary_load = pyo.Var(initialize=aux_load)  # kW
    m.fs.b14.auxilliary_load.fix()
    get_PP_costing(m.fs.b14, auxilliary_load_accounts,
                   m.fs.b14.auxilliary_load, 'kW', 6)

    # Accounts with STG,CTG output as the reference/scaling parameter
    # Case B31A Account 9 - Pg 501 rev 4 baseline report
    stg_ctg_accounts = ['11.8']
    m.fs.b15 = pyo.Block()
    # Obtain STG,CTG output in kW
    stg_ctg_op = 689800  # kW

    m.fs.b15.stg_ctg_output = pyo.Var(initialize=stg_ctg_op)  # kW
    m.fs.b15.stg_ctg_output.fix()
    get_PP_costing(m.fs.b15, stg_ctg_accounts,
                   m.fs.b15.stg_ctg_output, 'kW', 6)

    # Accounts with gas turbine power as the reference/scaling parameter
    # Exhibit 5-9
    gasturbine_accounts = ['14.1']
    m.fs.b16 = pyo.Block()
    # Obtain gas turbine power in kW
    gt_power = 477*1000  # kW

    m.fs.b16.gas_turbine_power = pyo.Var(initialize=gt_power)  # kW
    m.fs.b16.gas_turbine_power.fix()
    get_PP_costing(m.fs.b16, gasturbine_accounts,
                   m.fs.b16.gas_turbine_power, 'kW', 6)

    # Initialize costing
    costing_initialization(m.fs)
    assert degrees_of_freedom(m) == 0

    # Solve the model
    results = solver.solve(m, tee=True)
    assert results.solver.termination_condition == \
        pyo.TerminationCondition.optimal
    assert results.solver.status == pyo.SolverStatus.ok

    # Accounts with condenser duty as reference parameter
    assert pytest.approx(14.27, abs=0.1) \
        == sum(pyo.value(m.fs.b10.costing.total_plant_cost[ac])
               for ac in Condenser_duty_accounts)

    # Accounts with cooling tower duty as reference parameter
    assert pytest.approx(14.73, abs=0.2) \
        == sum(pyo.value(m.fs.b11.costing.total_plant_cost[ac])
               for ac in Cooling_tower_accounts)


@pytest.mark.skipif(solver is None, reason="Solver not available")
@pytest.mark.component
def test_flowsheet_costing(build_costing):
    m = build_costing
    # Build cost constraints
    build_flowsheet_cost_constraint(m)

    # Initialize costing
    costing_initialization(m.fs)
    assert degrees_of_freedom(m) == 0

    # Solve the model
    results = solver.solve(m, tee=True)
    assert results.solver.termination_condition == \
        pyo.TerminationCondition.optimal
    assert results.solver.status == pyo.SolverStatus.ok

    # Verify total plant costs
    assert pytest.approx(574.85, abs=0.1) == pyo.value(m.fs.flowsheet_cost)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES), and is copyright (c) 2018-2021
# by the software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia University
# Research Corporation, et al.  All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and
# license information.
#################################################################################
##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2020, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
Tests to make sure the NGFC example builds and solves correctly.
"""

__author__ = "Alex Noring"

import os

import pytest
import pyomo.environ as pyo
from pyomo.common.fileutils import this_file_dir

from idaes.power_generation.flowsheets.NGFC.NGFC_flowsheet import (
    build_power_island,
    build_reformer,
    set_power_island_inputs,
    set_reformer_inputs,
    scale_flowsheet,
    initialize_power_island,
    initialize_reformer,
    connect_reformer_to_power_island,
    SOFC_ROM_setup,
    add_SOFC_energy_balance,
    add_result_constraints)

from idaes.core import FlowsheetBlock
from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.core.util import get_solver
from idaes.core.util import model_serializer as ms

solver_available = pyo.SolverFactory('ipopt').available()
solver = get_solver()


@pytest.fixture(scope="module")
def m():
    m = pyo.ConcreteModel(name='NGFC without carbon capture')
    m.fs = FlowsheetBlock(default={"dynamic": False})

    build_power_island(m)
    build_reformer(m)
    set_power_island_inputs(m)
    set_reformer_inputs(m)
    connect_reformer_to_power_island(m)

    return m


@pytest.mark.component
def test_build(m):
    """Build NGFC and check for unit ops"""

    assert degrees_of_freedom(m) == 0

    assert hasattr(m.fs, 'anode')
    assert hasattr(m.fs.anode, 'heat_duty')
    assert hasattr(m.fs, 'cathode_heat')
    assert hasattr(m.fs.cathode_heat, 'heat_duty')
    assert hasattr(m.fs, 'reformer')
    assert hasattr(m.fs.reformer, 'heat_duty')
    assert hasattr(m.fs.reformer, 'deltaP')

    assert not m.fs.anode_mix.feed.flow_mol[0].fixed


@pytest.mark.integration
def test_initialize(m):
    scale_flowsheet(m)
    initialize_power_island(m)
    initialize_reformer(m)

    assert (pyo.value(m.fs.reformer.lagrange_mult[(0, 'H')]) ==
            pytest.approx(81235, 1e-3))
    assert (pyo.value(m.fs.prereformer.lagrange_mult[(0, 'H')]) ==
            pytest.approx(62743, 1e-3))
    assert (pyo.value(m.fs.anode.lagrange_mult[(0, 'H')]) ==
            pytest.approx(78300, 1e-3))

    assert (pyo.value(m.fs.bypass_rejoin.outlet.temperature[0]) ==
            pytest.approx(621.6, 1e-3))
    assert (pyo.value(m.fs.anode_hx.shell_outlet.temperature[0]) ==
            pytest.approx(826.5, 1e-3))
    assert (pyo.value(m.fs.cathode_hx.shell_outlet.temperature[0]) ==
            pytest.approx(475.8, 1e-3))


@pytest.mark.integration
def test_ROM(m):
    SOFC_ROM_setup(m)
    add_SOFC_energy_balance(m)
    add_result_constraints(m)

    assert degrees_of_freedom(m) == 0

    assert hasattr(m.fs, 'ROM_fuel_inlet_temperature')
    assert hasattr(m.fs, 'ROM_internal_reformation')
    assert m.fs.SOFC.current_density.fixed
    assert not m.fs.air_blower.inlet.flow_mol[0].fixed
    assert m.fs.SOFC.deltaT_cell.fixed
    assert hasattr(m.fs, 'stack_power')
    assert hasattr(m.fs, 'SOFC_energy_balance')
    assert not m.fs.anode.outlet.temperature[0].fixed
    assert hasattr(m.fs, 'CO2_emissions')


@pytest.mark.integration
def test_json_load(m):
    fname = os.path.join(os.path.join(os.path.dirname(this_file_dir()),
                                      "NGFC"), "NGFC_flowsheet_init.json.gz")

    ms.from_json(m, fname=fname)

    assert (pyo.value(m.fs.cathode.ion_outlet.flow_mol[0]) ==
            pytest.approx(1670.093, 1e-5))
    assert (pyo.value(m.fs.reformer_recuperator.area) ==
            pytest.approx(4512.56, 1e-3))
    assert (pyo.value(m.fs.anode.heat_duty[0]) ==
            pytest.approx(-672918626, 1e-5))
    assert (pyo.value(m.fs.CO2_emissions) ==
            pytest.approx(291.169, 1e-5))
    assert (pyo.value(m.fs.net_power) ==
            pytest.approx(659.879, 1e-5))





"""
Run all build steps for IDAES documentation.
"""
# standard libary
import argparse
import glob
import logging
import os
from pathlib import Path
import shutil
import subprocess
import sys

__author__ = "Dan Gunter"


_log = logging.getLogger("build_docs")
__h = logging.StreamHandler()
__h.setFormatter(logging.Formatter("%(levelname)s %(asctime)s - %(message)s"))
_log.addHandler(__h)
_log.setLevel(logging.ERROR)


class CommandError(Exception):
    def __init__(self, cmd, why):
        super().__init__(why)
        self.command = cmd
        self.message = why


def pipeline(*commands, **params):
    dry_run = params.get("dry_run", False)
    if dry_run:
        print_header("Dry run", first=True)
    first = True
    for cmd in commands:
        func = globals()[f"run_{cmd}"]
        if not dry_run:
            print_header(cmd, first=first)
        func(**params)
        first = False


def run_apidoc(clean=True, dry_run=False, **kwargs):
    """Run the sphinx-apidoc extension to build API docs.
    """
    if clean and not dry_run:
        # remove old docs
        print_status("Remove old docs")
        for old_rst in glob.glob("apidoc/*.rst"):
            os.unlink(old_rst)
    # run sphinx-apidoc
    print_status("Run sphinx-apidoc")
    os.environ["SPHINX_APIDOC_OPTIONS"] = "members,ignore-module-all,noindex"
    _run(
        "apidoc",
        [
            "sphinx-apidoc",
            "--module-first",
            "--output-dir",
            "apidoc",
            "../idaes",
            "../idaes/*tests*",
        ],
        60,
        dry_run
    )
    if not dry_run:
        postprocess_apidoc(Path("apidoc"))


def postprocess_apidoc(root):
    """Perform postprocessing on generated apidoc files
    """
    # Remove :noindex: from all entries in given modules
    remove_noindex = ["idaes.dmf"]
    for module in remove_noindex:
        module_path = root / (module + ".rst")
        _log.debug(f"Looking for :noindex: in {module_path}")
        lines, has_noindex = [], False
        with module_path.open("r") as input_file:
            for line in input_file:
                if ":noindex:" in line:
                    has_noindex = True
                else:
                    lines.append(line)
        if has_noindex:
            _log.debug(f"Removing :noindex: from {module_path}")
            with module_path.open("w") as output_file:
                for line in lines:
                    output_file.write(line)


def run_html(clean=True, vb=0, timeout=0, dry_run=False, **kwargs):
    """Run sphinx-build to create HTML.
    """
    build_dir = "build"
    output_file = "sphinx-errors.txt"
    if not dry_run:
        if clean:
            if os.path.exists(build_dir):
                print_status(f"Remove build dir '{build_dir}'")
                shutil.rmtree(build_dir)
        if os.path.exists(output_file):
            os.unlink(output_file)
    # run
    print_status("Run sphinx-build")
    if vb > 0:
        verbosity = "-" + "v" * vb
    else:
        verbosity = "-q"
    _run(
        "html",
        ["sphinx-build", "-M", "html", ".", build_dir, "-w", output_file] + [verbosity],
        timeout,
        dry_run
    )
    if dry_run:
        return
    # check output file
    print_status("Check output file")
    num_lines = num_warnings = num_errors = 0
    if os.path.exists(output_file):
        # Count lines, warnings and errors
        with open(output_file) as outf:
            for line in outf:
                num_lines += 1
                num_warnings = num_warnings + 1 if "WARNING: " in line else num_warnings
                num_errors = num_errors + 1 if "ERROR: " in line else num_errors
    if num_lines > 0:
        raise CommandError(
            "html",
            f"sphinx-build had {num_warnings} warnings, {num_errors} errors in {num_lines} lines of output\n"
            f"These will cause tests to fail\n"
            f"See file '{output_file}' for details",
        )


def _run(what, args, timeout, not_really):
    """Run a command with a timeout.
    """
    command_line = ' '.join(args)
    if not_really:
        print(f"Command: {command_line}")
        return
    _log.debug(f"command={command_line}")
    try:
        proc = subprocess.Popen(args)
    except Exception as err:
        exe = args[0]
        raise CommandError(what, f"Could not run '{exe}':\n{err}")
    try:
        proc.wait(timeout)
    except subprocess.TimeoutExpired:
        raise CommandError(what, f"Timed out after {timeout} seconds")


def print_header(msg: str, first: bool = False):
    if not first:
        print()
    print(f"=== {msg}")


def print_status(msg: str):
    print(f"--- {msg}")


def print_error(cmd, msg):
    lines = msg.split("\n")
    print()
    print(f"*** ERROR in '{cmd}'")
    print("***")
    for line in lines:
        print(f"*** {line}")
    print("***")
    print()


def main() -> int:
    """Entry point for this module.
    """
    prs = argparse.ArgumentParser(description=__doc__.strip())
    prs.add_argument(
        "--dirty",
        action="store_true",
        help="Do not clean files before running commands",
    )
    prs.add_argument(
        "-d",
        "--dry-run",
        action="store_true",
        dest="dry_run",
        help="Show commands to be executed",
    )
    prs.add_argument(
        "-t",
        "--timeout",
        dest="timeout",
        help="Timeout (in seconds) for sphinx-build (default=180)",
        default=180,
        type=int,
    )
    prs.add_argument(
        "-v",
        "--verbose",
        action="count",
        dest="vb",
        default=0,
        help="Verbosity for `sphinx-build`; also some debug information from this program",
    )
    args = prs.parse_args()
    verbosity = 0
    if args.vb > 0:
        _log.setLevel(logging.DEBUG)
        verbosity = args.vb
    try:
        pipeline(
            "apidoc", "html", clean=not args.dirty, vb=verbosity, timeout=args.timeout,
            dry_run=args.dry_run
        )
    except CommandError as err:
        print_error(err.command, err.message)
        return -1
    if not args.dry_run:
        print_header("SUCCESS")
    return 0


if __name__ == "__main__":
    sys.exit(main())
